# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from flask import Blueprint, Flask, jsonify, request
from werkzeug.datastructures import Headers

from rucio import version
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, check_accept_header_wrapper_flask, response_headers

if TYPE_CHECKING:
    from typing import Optional

    from rucio.web.rest.flaskapi.v1.types import HeadersType


class Ping(ErrorHandlingMethodView):

    def get_headers(self) -> "Optional[HeadersType]":
        headers = Headers()
        headers.set('Access-Control-Allow-Origin', request.environ.get('HTTP_ORIGIN'))
        headers.set('Access-Control-Allow-Headers', request.environ.get('HTTP_ACCESS_CONTROL_REQUEST_HEADERS'))
        headers.set('Access-Control-Allow-Methods', '*')
        headers.set('Access-Control-Allow-Credentials', 'true')

        headers.set('Cache-Control', 'no-cache, no-store, max-age=0, must-revalidate')
        headers.add('Cache-Control', 'post-check=0, pre-check=0')
        headers.set('Pragma', 'no-cache')
        return headers

    @check_accept_header_wrapper_flask(['application/json'])
    def get(self):
        """
        ---
        summary: Ping
        description: Ping the server and get data about it.
        tags:
          - Ping
        responses:
          200:
            description: OK
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    version:
                      description: The server version.
                      type: string
          406:
            description: Not acceptable
        """
        headers = self.get_headers()
        response = jsonify(version=version.version_string())
        response.headers.extend(headers)
        return response


def blueprint(standalone=False, with_doc=False):
    bp = Blueprint('ping', __name__, url_prefix='/' if standalone else '/ping')

    ping_view = Ping.as_view('ping')
    if not with_doc:
        # rule without trailing slash needs to be added before rule with trailing slash
        bp.add_url_rule('', view_func=ping_view, methods=['get', ])
    bp.add_url_rule('/', view_func=ping_view, methods=['get', ])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation to add the prefix """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint(with_doc=True))
    return doc_app



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from collections import namedtuple
from typing import Literal, get_args

from rucio.common.config import config_get_bool

"""
Constants.

"""

RESERVED_KEYS = ['scope', 'name', 'account', 'did_type', 'is_open', 'monotonic', 'obsolete', 'complete',
                 'availability', 'suppressed', 'bytes', 'length', 'md5', 'adler32', 'rule_evaluation_action',
                 'rule_evaluation_required', 'expired_at', 'deleted_at', 'created_at', 'updated_at']
# collection_keys =
# file_keys =

KEY_TYPES = ['ALL', 'COLLECTION', 'FILE', 'DERIVED']
# all(container, dataset, file), collection(dataset or container), file, derived(compute from file for collection)

SCHEME_MAP = {'srm': ['srm', 'gsiftp'],
              'gsiftp': ['srm', 'gsiftp'],
              'https': ['https', 'davs', 'srm+https', 'cs3s'],
              'davs': ['https', 'davs', 'srm+https', 'cs3s'],
              'srm+https': ['https', 'davs', 'srm+https', 'cs3s'],
              'cs3s': ['https', 'davs', 'srm+https', 'cs3s'],
              'root': ['root'],
              'scp': ['scp'],
              'rsync': ['rsync'],
              'rclone': ['rclone']}
if config_get_bool('transfers', 'srm_https_compatibility', raise_exception=False, default=False):
    SCHEME_MAP['srm'].append('https')
    SCHEME_MAP['https'].append('srm')
    SCHEME_MAP['srm'].append('davs')
    SCHEME_MAP['davs'].append('srm')

SORTING_ALGORITHMS_LITERAL = Literal['geoip', 'custom_table', 'random']
SORTING_ALGORITHMS = list(get_args(SORTING_ALGORITHMS_LITERAL))

SUPPORTED_PROTOCOLS_LITERAL = Literal['gsiftp', 'srm', 'root', 'davs', 'http', 'https', 'file', 'storm', 'srm+https', 'scp', 'rsync', 'rclone', 'magnet']
SUPPORTED_PROTOCOLS: list[str] = list(get_args(SUPPORTED_PROTOCOLS_LITERAL))

RSE_SUPPORTED_PROTOCOL_DOMAINS_LITERAL = Literal['ALL', 'LAN', 'WAN']

RSE_BASE_SUPPORTED_PROTOCOL_OPERATIONS_LITERAL = Literal['read', 'write', 'delete']
RSE_BASE_SUPPORTED_PROTOCOL_OPERATIONS: list[str] = list(get_args(RSE_BASE_SUPPORTED_PROTOCOL_OPERATIONS_LITERAL))

RSE_ALL_SUPPORTED_PROTOCOL_OPERATIONS_LITERAL = Literal[RSE_BASE_SUPPORTED_PROTOCOL_OPERATIONS_LITERAL, 'third_party_copy_read', 'third_party_copy_write']
RSE_ALL_SUPPORTED_PROTOCOL_OPERATIONS: list[str] = list(get_args(RSE_ALL_SUPPORTED_PROTOCOL_OPERATIONS_LITERAL))

FTS_STATE = namedtuple('FTS_STATE', ['SUBMITTED', 'READY', 'ACTIVE', 'FAILED', 'FINISHED', 'FINISHEDDIRTY', 'NOT_USED',
                                     'CANCELED'])('SUBMITTED', 'READY', 'ACTIVE', 'FAILED', 'FINISHED', 'FINISHEDDIRTY',
                                                  'NOT_USED', 'CANCELED')

FTS_COMPLETE_STATE = namedtuple('FTS_COMPLETE_STATE', ['OK', 'ERROR'])('Ok', 'Error')

# https://gitlab.cern.ch/fts/fts3/-/blob/master/src/db/generic/Job.h#L41
FTS_JOB_TYPE = namedtuple('FTS_JOB_TYPE', ['MULTIPLE_REPLICA', 'MULTI_HOP', 'SESSION_REUSE', 'REGULAR'])('R', 'H', 'Y', 'N')


# Messages constants

MAX_MESSAGE_LENGTH = 4000


@enum.unique
class SuspiciousAvailability(enum.Enum):
    ALL = 0
    EXIST_COPIES = 1
    LAST_COPY = 2


@enum.unique
class ReplicaState(enum.Enum):
    # From rucio.db.sqla.constants, update that file at the same time as this
    AVAILABLE = 'A'
    UNAVAILABLE = 'U'
    COPYING = 'C'
    BEING_DELETED = 'B'
    BAD = 'D'
    TEMPORARY_UNAVAILABLE = 'T'


@enum.unique
class HermesService(str, enum.Enum):
    """
    The services supported by Hermes2.
    """
    INFLUX = "INFLUX"
    ELASTIC = "ELASTIC"
    EMAIL = "EMAIL"
    ACTIVEMQ = "ACTIVEMQ"


class RseAttr:

    """
    List of functional RSE attributes.

    This class acts as a namespace containing all RSE attributes referenced in
    the Rucio source code. Setting them affects Rucio's behaviour in some way.
    """

    ARCHIVE_TIMEOUT = 'archive_timeout'
    ASSOCIATED_SITES = 'associated_sites'
    AUTO_APPROVE_BYTES = 'auto_approve_bytes'
    AUTO_APPROVE_FILES = 'auto_approve_files'
    BITTORRENT_TRACKER_ADDR = 'bittorrent_tracker_addr'
    BLOCK_MANUAL_APPROVAL = 'block_manual_approval'
    COUNTRY = 'country'
    DECOMMISSION = 'decommission'
    DEFAULT_ACCOUNT_LIMIT_BYTES = 'default_account_limit_bytes'
    FTS = 'fts'
    GLOBUS_ENDPOINT_ID = 'globus_endpoint_id'
    GREEDYDELETION = 'greedyDeletion'
    IS_OBJECT_STORE = 'is_object_store'
    LFN2PFN_ALGORITHM = 'lfn2pfn_algorithm'
    MAXIMUM_PIN_LIFETIME = 'maximum_pin_lifetime'
    MULTIHOP_TOMBSTONE_DELAY = 'multihop_tombstone_delay'
    NAMING_CONVENTION = 'naming_convention'
    OIDC_BASE_PATH = 'oidc_base_path'
    OIDC_SUPPORT = 'oidc_support'
    PHYSGROUP = 'physgroup'
    QBITTORRENT_MANAGEMENT_ADDRESS = 'qbittorrent_management_address'
    RESTRICTED_READ = 'restricted_read'
    RESTRICTED_WRITE = 'restricted_write'
    RULE_APPROVERS = 'rule_approvers'
    S3_URL_STYLE = 's3_url_style'
    SIGN_URL = 'sign_url'
    SIMULATE_MULTIRANGE = 'simulate_multirange'
    SITE = 'site'
    SKIP_UPLOAD_STAT = 'skip_upload_stat'
    SOURCE_FOR_TOTAL_SPACE = 'source_for_total_space'
    SOURCE_FOR_USED_SPACE = 'source_for_used_space'
    STAGING_BUFFER = 'staging_buffer'
    STAGING_REQUIRED = 'staging_required'
    STRICT_COPY = 'strict_copy'
    TOMBSTONE_DELAY = 'tombstone_delay'
    TYPE = 'type'
    USE_IPV4 = 'use_ipv4'
    VERIFY_CHECKSUM = 'verify_checksum'

    # The following RSE attributes are exclusively used in the permission layer
    # and are likely VO-specific.

    BLOCK_MANUAL_APPROVE = 'block_manual_approve'
    CMS_TYPE = 'cms_type'
    DEFAULT_LIMIT_FILES = 'default_limit_files'
    QUOTA_APPROVERS = 'quota_approvers'
    RULE_DELETERS = 'rule_deleters'


# Literal types to allow overloading of functions with RSE attributes in their signature.
# RSE attributes are encoded via the BooleanString decorator as VARCHAR in the database,
# but they are used as either bool or string in the code.
# This is only determined at runtime, so for static type checking
# we need to manually specify which attrs are string and which are bool.
# In future, we could refactor RseAttr to avoid code duplication.
RSE_ATTRS_STR = Literal[
    'archive_timeout',
    'associated_sites',
    'bittorrent_tracker_addr',
    'country',
    'decommission',
    'default_account_limit_bytes',
    'fts',
    'globus_endpoint_id',
    'lfn2pfn_algorithm',
    'maximum_pin_lifetime',
    'multihop_tombstone_delay',
    'naming_convention',
    'oidc_base_path',
    'oidc_support'
    'physgroup',
    'qbittorrent_management_address'
    'rule_approvers',
    's3_url_style',
    'simulate_multirange',
    'site',
    'source_for_total_space',
    'source_for_used_space',
    'staging_buffer',
    'tombstone_delay',
    'type'
]

RSE_ATTRS_BOOL = Literal[
    'auto_approve_bytes',
    'auto_approve_files',
    'block_manual_approval',
    'greedyDeletion',
    'is_object_store',
    'restricted_read',
    'restricted_write',
    'skip_upload_stat',
    'staging_required',
    'strict_copy',
    'use_ipv4',
    'verify_checksum'
]

SUPPORTED_SIGN_URL_SERVICES_LITERAL = Literal['gcs', 's3', 'swift']
SUPPORTED_SIGN_URL_SERVICES = list(get_args(SUPPORTED_SIGN_URL_SERVICES_LITERAL))



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import uuid
from typing import TYPE_CHECKING, Any, Optional

from rucio.db.sqla.constants import RequestState
from rucio.transfertool.transfertool import TransferStatusReport, Transfertool, TransferToolBuilder

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence

    from rucio.common.types import LoggerFunction
    from rucio.core.request import DirectTransfer
    from rucio.db.sqla.session import Session


class MockTransferStatusReport(TransferStatusReport):

    supported_db_fields = [
        'state',
        'external_id'
    ]

    def __init__(self, request_id: str, external_id: str):
        super().__init__(request_id)
        self.state = RequestState.DONE
        self.external_id = external_id

    def initialize(self, session: "Session", logger: "LoggerFunction" = logging.log) -> None:
        pass

    def get_monitor_msg_fields(self, session: "Session", logger: "LoggerFunction" = logging.log) -> dict[str, Any]:
        return {}


class MockTransfertool(Transfertool):
    """
    Mock implementation of a Rucio transfertool

    This is not actually used anywhere at the moment
    """

    external_name = 'mock'
    required_rse_attrs = ()
    supported_schemes = {'mock', 'file'}

    def __init__(self, external_host: str, logger: "LoggerFunction" = logging.log):
        super(MockTransfertool, self).__init__(external_host, logger)

    @classmethod
    def submission_builder_for_path(
        cls,
        transfer_path: list["DirectTransfer"],
        logger: "LoggerFunction" = logging.log
    ) -> tuple[list["DirectTransfer"], "TransferToolBuilder"]:
        return transfer_path, TransferToolBuilder(cls, external_host='Mock Transfertool')

    def group_into_submit_jobs(self, transfers: "Iterable[list[DirectTransfer]]") -> list[dict[str, Any]]:
        return [{'transfers': list(itertools.chain.from_iterable(transfers)), 'job_params': {}}]

    def submit(self, transfers: "Iterable[DirectTransfer]", job_params: dict[str, str], timeout: Optional[int] = None) -> str:
        return str(uuid.uuid1())

    def bulk_query(self, requests_by_eid: "Mapping[str, Mapping[str, Any]]", timeout: Optional[int] = None) -> dict[str, dict[str, MockTransferStatusReport]]:
        response = {}
        for transfer_id, requests in requests_by_eid.items():
            for request_id in requests:
                response.setdefault(transfer_id, {})[request_id] = MockTransferStatusReport(request_id, transfer_id)
        return response

    def cancel(self, transfer_ids: 'Sequence[str]', timeout: Optional[int] = None) -> bool:
        return True

    def update_priority(self, transfer_id: str, priority: int, timeout: Optional[int] = None) -> bool:
        return True



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import functools
import itertools
import json
import logging
import re
import sys
from traceback import format_tb
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, get_args

from rucio.common.config import config_get, config_get_bool

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Mapping, Sequence
    from logging import LogRecord, _SysExcInfoType

    from _typeshed import OptExcInfo
    from flask import Flask


# Mapping from ECS field paths
# https://www.elastic.co/guide/en/ecs-logging/overview/current/intro.html#_field_mapping
# https://www.elastic.co/guide/en/ecs/8.5/ecs-field-reference.html
# to python log record attributes:
# https://docs.python.org/3/library/logging.html#logrecord-attributes
ECS_FIELDS = Literal[
    '@timestamp',
    'message',
    'log.level',
    'log.origin.function',
    'log.origin.file.line',
    'log.origin.file.name',
    'log.logger',
    'process.pid',
    'process.name',
    'process.thread.id',
    'process.thread.name'
]

LOG_RECORDS = Literal[
    'asctime',
    'message',
    'levelname',
    'funcName',
    'lineno',
    'filename',
    'name',
    'process',
    'processName',
    'thread',
    'threadName'
]

BUILTIN_FIELDS: tuple[tuple[ECS_FIELDS, LOG_RECORDS], ...] = tuple((x, y) for x, y in zip(get_args(ECS_FIELDS), get_args(LOG_RECORDS)))
ECS_TO_LOG_RECORD_MAP: dict[ECS_FIELDS, LOG_RECORDS] = dict(BUILTIN_FIELDS)
LOG_RECORD_TO_ECS_MAP: dict[LOG_RECORDS, ECS_FIELDS] = dict((f[1], f[0]) for f in BUILTIN_FIELDS)


def _json_serializable(obj: Any) -> Union[dict[Any, Any], str]:
    try:
        return obj.__dict__
    except AttributeError:
        return str(obj)


def _navigate_path(obj: Any, path: 'Sequence[str]') -> Optional[Any]:
    """
    Traverse the path in the given object either via attributes or via dict-like subscriptions.
    Returns the found value; None if navigation fails

    For example, for an input
      obj = request  # flask "request" object https://flask.palletsprojects.com/en/2.1.x/api/#flask.Request
      path = ['headers', 'X-Rucio-Auth-Token']
    returns the value found in
      request.headers['X-Rucio-Auth-Token']
    """
    value = obj
    i = 0
    while value and i < len(path):
        p = path[i]
        try:
            value = getattr(value, p)
        except AttributeError:
            try:
                # Allow integers for access into arrays
                p = int(p)
            except ValueError:
                pass
            try:
                value = value[p]
            except (TypeError, KeyError):
                value = None
        i += 1
    if value is obj:
        return None
    return value


def _unflatten_dict(dictionary: dict[str, Any]) -> dict[str, Any]:
    """
    Transform a dict of the form
    {'a.b.c': value1, 'a.b.d': value2, 'z': value3}
    into
    {'a': {'b': {'c': value1, 'd': value2}}, 'z': value3}

    On incompatible input keys (for example: 'a.b.c', 'a', 'a.d'), the last key wins
    """
    ret = {}
    for k, v in dictionary.items():
        path = k.split('.')
        d = ret
        i = 0
        while i < len(path) - 1:
            existing_v = d.get(path[i])
            if isinstance(existing_v, dict):
                d = existing_v
            else:
                d[path[i]] = {}
                d = d[path[i]]
            i += 1
        if i == len(path) - 1:
            d[path[i]] = v
    return ret


def _get_request_data(request_path: 'Sequence[str]') -> "Callable[[LogDataSource, LogRecord], Iterator[tuple[str, Optional[Any]]]]":
    """
    Returns a function which, when called, will resolve the value
    in the flask request object at request_path
    """

    # The import fails if imported inside a client due to rsemanager.
    # TODO: move to top of file once we got rid of/refactored rsemanager
    from flask import has_request_context, request

    def _request_data_formatter(record_formatter: "LogDataSource", record: "LogRecord") -> 'Iterator[tuple[str, Optional[Any]]]':
        value = None
        if has_request_context() and request_path:
            value = _navigate_path(request, request_path)
        yield record_formatter.ecs_fields[0], str(value) if value is not None else None

    return _request_data_formatter


def _get_record_attribute(attribute: str) -> "Callable[[LogDataSource, LogRecord], Iterator[tuple[str, Optional[Any]]]]":
    """
    Returns a function which, when called, will generate the value of the desired attribute from
    the record passed in argument.
    """

    def _record_attribute_formatter(record_formatter: "LogDataSource", record: "LogRecord") -> 'Iterator[tuple[str, Optional[Any]]]':
        value = None
        try:
            value = getattr(record, attribute)
        except AttributeError:
            pass
        yield record_formatter.ecs_fields[0], value

    return _record_attribute_formatter


def _timestamp_formatter(record_formatter: "LogDataSource", record: "LogRecord") -> 'Iterator[tuple[str, Optional[Any]]]':
    """
    Format a timestamp
    """
    yield record_formatter.ecs_fields[0], datetime.datetime.utcfromtimestamp(record.created).isoformat(timespec='milliseconds') + 'Z'


def _ecs_field_to_record_attribute(field_name: Union[ECS_FIELDS, str]) -> Union[LOG_RECORDS, str]:
    """
    Sanitize the path-like field name into a symbol which can be the name of an object attribute.
    """
    record = ECS_TO_LOG_RECORD_MAP.get(field_name)  # type: ignore
    if record:
        return record
    return field_name.replace('-', '_').replace('.', '_')


class LogDataSource:
    """
    Represents one log data source and allows to format it into one or more json fields
    """
    def __init__(
            self,
            ecs_fields: tuple[str, ...],
            formatter: "Optional[Callable[[LogDataSource, LogRecord], Iterator[tuple[str, Optional[Any]]]]]" = None,
            dst_record_attr: Optional[str] = None
    ):
        self.ecs_fields = ecs_fields
        self._formatter = formatter
        self._dst_record_attr = dst_record_attr

    def __hash__(self):
        return hash(self.ecs_fields)

    def __eq__(self, other: Any):
        if not other or not isinstance(other, self.__class__):
            return False
        return self.ecs_fields == other.ecs_fields

    def __str__(self):
        return self.__class__.__name__ + '(' + ', '.join(self.ecs_fields) + ')'

    def format(self, record: "LogRecord") -> Optional['Iterator[tuple[str, Any]]']:
        if not self._formatter:
            return
        for field_name, field_value in self._formatter(self, record):
            if self._dst_record_attr:
                setattr(record, self._dst_record_attr, field_value)
            yield field_name, field_value


class MessageLogDataSource(LogDataSource):
    def __init__(self):
        super().__init__(
            ecs_fields=('message', 'error.type', 'error.message', 'error.stack_trace'),
            formatter=None,
        )

    @staticmethod
    def _get_exc_info(record: "LogRecord") -> Optional[Union["OptExcInfo", "_SysExcInfoType"]]:
        exc_info = record.exc_info
        if not exc_info:
            return None
        if isinstance(exc_info, bool):
            exc_info = sys.exc_info()
        if isinstance(exc_info, (list, tuple)):
            return exc_info
        return None

    def format(self, record: "LogRecord") -> 'Iterator[tuple[str, Optional[str]]]':
        exc_info = self._get_exc_info(record)
        message = record.getMessage()
        error_type, error_message, stack_trace = None, None, None
        if exc_info:
            error_type = exc_info[0].__name__ if exc_info[0] else None
            error_message = str(exc_info[1]) if exc_info[1] else None
            stack_trace = "".join(format_tb(record.exc_info[2])) or None if exc_info[2] else None
            if not stack_trace:
                stack_trace = str(getattr(record, "stack_info", '')) or None

        # Set the message into the record field
        s = message
        if error_message:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + error_message
        if stack_trace:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + stack_trace
        record.message = s

        yield from zip(self.ecs_fields, (record.message, error_type, error_message, stack_trace))


class ConstantStrDataSource(LogDataSource):
    """
    Prints a constant string for the given ECS field.
    """

    def __init__(self, ecs_field: ECS_FIELDS, _str: str):
        log_record = ECS_TO_LOG_RECORD_MAP.get(ecs_field, None)
        self._str = _str

        def _formatter(data_source: LogDataSource, record: "LogRecord") -> 'Iterator[tuple[str, str]]':
            yield self.ecs_fields[0], self._str

        super().__init__(ecs_fields=(ecs_field,), formatter=_formatter, dst_record_attr=log_record)


class RucioFormatter(logging.Formatter):
    """
    The formatter should be a drop-in replacement to the python builtin
    formatter, with two additional additions:
    - it can output directly to json
    - it can include data from the flask 'request' object into the format

    When the logger writes to a json format, it tries to respect the
    Elastic Common Schema (ECS) specification, but without getting too
    strict about it.

    When the format string contains a dot-separated "path" starting with
    `http.request.`, the rucio formatter will try to extract the given
    path from the flask `request` object.
    """

    def __init__(
            self,
            fmt: Optional[str] = None,
            validate: Optional[bool] = None,
            output_json: bool = False,
            additional_fields: Optional['Mapping[ECS_FIELDS, str]'] = None
    ):
        _kwargs = {}
        if validate is not None:
            _kwargs["validate"] = validate

        data_sources: dict[str, LogDataSource] = dict(
            (ecs_field, LogDataSource((ecs_field,), formatter=_get_record_attribute(log_record)))
            for ecs_field, log_record in BUILTIN_FIELDS
        )
        data_sources.update({
            '@timestamp': LogDataSource(('@timestamp',), formatter=_timestamp_formatter),
            'message': MessageLogDataSource(),  # ('message', 'error.type', 'error.message', 'error.stack_trace'),
        })
        data_sources.update(
            (ecs_field, LogDataSource((ecs_field,),
                                      dst_record_attr=_ecs_field_to_record_attribute(ecs_field),
                                      formatter=_get_request_data(request_path=request_path.split('.'))))
            for ecs_field, request_path in (
                ('client.account.name', 'headers.X-Rucio-Account'),  # this field is rucio-specific, not from the ECS specification
                ('network.forwarded_ip', 'access_route.0'),
                ('source.ip', 'remote_addr'),
                ('url.full', 'url'),
                ('user_agent.original', 'user_agent'),
            )
        )
        if additional_fields:
            data_sources.update({
                ecs_field: ConstantStrDataSource(ecs_field, field_value)
                for ecs_field, field_value in additional_fields.items()
            })

        self._desired_data_sources = []
        if fmt:
            # extract of field1, field2 from the printf format-string "%(field1)s %(field2)i"
            # Allow simple path-like structures in fields (words separated with dots):
            # - http.request.headers.X-Rucio-Auth-Token
            # - http.request.url
            _format_string_fields = set(t[0] for t in re.findall(r'%\((\w+(.\w+(-\w+)*)*)\)', fmt))

            for field_name in _format_string_fields:
                data_source = data_sources.get(LOG_RECORD_TO_ECS_MAP.get(field_name, field_name), None)

                if '.' in field_name:
                    dst_record_attr = _ecs_field_to_record_attribute(field_name)
                    fmt = fmt.replace(f'%({field_name})', f'%({dst_record_attr})')
                    if field_name.startswith('http.request.'):
                        path = field_name.replace('http.request.', '', 1).split('.')
                        data_source = LogDataSource((field_name,), dst_record_attr=dst_record_attr, formatter=_get_request_data(path))
                elif not data_source:
                    data_source = LogDataSource((field_name,), formatter=_get_record_attribute(field_name))

                if data_source:
                    self._desired_data_sources.append(data_source)
        else:
            self._desired_data_sources = [data_sources['message']]

        self.output_json = output_json
        super().__init__(fmt=fmt, style='%', **_kwargs)

    def format(self, record: "LogRecord") -> str:
        json_record = dict(itertools.chain.from_iterable(f.format(record) for f in self._desired_data_sources))  # type: ignore
        if self.output_json:
            return self._to_json(_unflatten_dict(json_record))
        else:
            return super().format(record)

    @staticmethod
    def _to_json(record: dict[str, Any]) -> str:
        try:
            return json.dumps(record, default=_json_serializable)
        except (TypeError, ValueError, OverflowError):
            try:
                return json.dumps(record)
            except (TypeError, ValueError, OverflowError):
                return '{}'


def rucio_log_formatter(process_name: Optional[str] = None) -> RucioFormatter:
    config_logformat = config_get('common', 'logformat', raise_exception=False, default='%(asctime)s\t%(name)s\t%(process)d\t%(levelname)s\t%(message)s')
    output_json = config_get_bool('common', 'logjson', default=False)
    additional_fields: 'Mapping[ECS_FIELDS, str]' = {}
    if process_name:
        additional_fields['process.name'] = process_name
    return RucioFormatter(fmt=config_logformat, output_json=output_json, additional_fields=additional_fields)


def setup_logging(application: Optional["Flask"] = None, process_name: Optional[str] = None) -> None:
    """
    Configures the logging by setting the output stream to stdout and
    configures log level and log format.
    """
    config_loglevel = getattr(logging, config_get('common', 'loglevel', raise_exception=False, default='DEBUG').upper())

    stdouthandler = logging.StreamHandler(stream=sys.stdout)
    stdouthandler.setFormatter(rucio_log_formatter(process_name=process_name))
    stdouthandler.setLevel(config_loglevel)
    logging.basicConfig(level=config_loglevel, handlers=[stdouthandler])

    if application:
        application.logger.addHandler(stdouthandler)


def formatted_logger(innerfunc: 'Callable', formatstr: str = "%s") -> 'Callable':
    """
    Decorates the passed function, formatting log input by
    the passed formatstr. The format string must always include a %s.

    :param innerfunc: function to be decorated. Must take (level, msg) arguments.
    :param formatstr: format string with %s as placeholder.
    """
    @functools.wraps(innerfunc)
    def log_format(level: int, msg: object, *args, **kwargs) -> 'Callable':
        return innerfunc(level, formatstr % msg, *args, **kwargs)
    return log_format



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os

from rucio.common import exception
from rucio.common.checksum import PREFERRED_CHECKSUM
from rucio.common.config import get_config_dirs
from rucio.common.utils import execute
from rucio.rse.protocols import protocol


def load_conf_file(file_name):
    config_dir = next(filter(lambda d: os.path.exists(os.path.join(d, file_name)), get_config_dirs()))
    with open(os.path.join(config_dir, file_name)) as f:
        return json.load(f)


class Default(protocol.RSEProtocol):
    """ Implementing access to RSEs using the rclone protocol."""

    def __init__(self, protocol_attr, rse_settings, logger=logging.log):
        """ Initializes the object with information about the referred RSE.

            :param props: Properties derived from the RSE Repository
        """
        super(Default, self).__init__(protocol_attr, rse_settings, logger=logger)
        if len(rse_settings['protocols']) == 1:
            raise exception.RucioException('rclone initialization requires at least one other protocol defined on the RSE. (from ssh, sftp, posix, webdav)')
        self.scheme = self.attributes['scheme']
        setuprclone = False
        for protocols in reversed(rse_settings['protocols']):
            if protocol_attr['impl'] == protocols['impl']:
                continue
            else:
                setuprclone = self.setuphostname(protocols)
                if setuprclone:
                    break

        if not setuprclone:
            raise exception.RucioException('rclone could not be initialized.')
        self.logger = logger

    def setuphostname(self, protocols):
        """ Initializes the rclone object with information about protocols in the referred RSE.

            :param protocols: Protocols in the RSE
        """
        if protocols['scheme'] in ['scp', 'rsync', 'sftp']:
            self.hostname = 'ssh_rclone_rse'
            self.host = protocols['hostname']
            self.port = str(protocols['port'])
            if protocols['extended_attributes'] is not None and 'user' in list(protocols['extended_attributes'].keys()):
                self.user = protocols['extended_attributes']['user']
            else:
                self.user = None
            try:
                data = load_conf_file('rclone-init.cfg')
                key_file = data[self.host + '_ssh']['key_file']
            except KeyError:
                self.logger(logging.ERROR, 'rclone.init: rclone-init.cfg:- Field value missing for "{}_ssh: key_file"'.format(self.host))
                return False
            try:
                if self.user:
                    cmd = 'rclone config create {0} sftp host {1} user {2} port {3} key_file {4}'.format(self.hostname, self.host, self.user, str(self.port), key_file)
                    self.logger(logging.DEBUG, 'rclone.init: cmd: {}'.format(cmd))
                    status, out, err = execute(cmd)
                    if status:
                        return False
                else:
                    cmd = 'rclone config create {0} sftp host {1} port {2} key_file {3}'.format(self.hostname, self.host, str(self.port), key_file)
                    self.logger(logging.DEBUG, 'rclone.init: cmd: {}'.format(cmd))
                    status, out, err = execute(cmd)
                    if status:
                        return False
            except Exception as e:
                raise exception.ServiceUnavailable(e)

        elif protocols['scheme'] == 'file':
            self.hostname = '%s_rclone_rse' % (protocols['scheme'])
            self.host = 'localhost'
            try:
                cmd = 'rclone config create {0} local'.format(self.hostname)
                self.logger(logging.DEBUG, 'rclone.init: cmd: {}'.format(cmd))
                status, out, err = execute(cmd)
                if status:
                    return False
            except Exception as e:
                raise exception.ServiceUnavailable(e)

        elif protocols['scheme'] in ['davs', 'https']:
            self.hostname = '%s_rclone_rse' % (protocols['scheme'])
            self.host = protocols['hostname']
            url = '%s://%s:%s%s' % (protocols['scheme'], protocols['hostname'], str(protocols['port']), protocols['prefix'])
            try:
                data = load_conf_file('rclone-init.cfg')
                bearer_token = data[self.host + '_webdav']['bearer_token']
            except KeyError:
                self.logger(logging.ERROR, 'rclone.init: rclone-init.cfg:- Field value missing for "{}_webdav: bearer_token"'.format(self.host))
                return False
            try:
                cmd = 'rclone config create {0} webdav url {1} vendor other bearer_token {2}'.format(self.hostname, url, bearer_token)
                self.logger(logging.DEBUG, 'rclone.init: cmd: {}'.format(cmd))
                status, out, err = execute(cmd)
                if status:
                    return False
            except Exception as e:
                raise exception.ServiceUnavailable(e)

        else:
            self.logger(logging.DEBUG, 'rclone.init: {} protocol impl not supported by rucio rclone'.format(protocols['impl']))
            return False

        return True

    def path2pfn(self, path):
        """
            Returns a fully qualified PFN for the file referred by path.

            :param path: The path to the file.

            :returns: Fully qualified PFN.

        """
        self.logger(logging.DEBUG, 'rclone.path2pfn: path: {}'.format(path))
        if not path.startswith('rclone://'):
            return '%s://%s/%s' % (self.scheme, self.host, path)
        else:
            return path

    def exists(self, pfn):
        """ Checks if the requested file is known by the referred RSE.

            :param pfn: Physical file name

            :returns: True if the file exists, False if it doesn't

            :raise  ServiceUnavailable
        """
        self.logger(logging.DEBUG, 'rclone.exists: pfn: {}'.format(pfn))
        try:
            path = self.pfn2path(pfn)
            cmd = 'rclone lsf %s:%s' % (self.hostname, path)
            self.logger(logging.DEBUG, 'rclone.exists: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status:
                return False
        except Exception as e:
            raise exception.ServiceUnavailable(e)

        return True

    def stat(self, path):
        """
        Returns the stats of a file.

        :param path: path to file

        :raises ServiceUnavailable: if some generic error occurred in the library.

        :returns: a dict with two keys, filesize and an element of GLOBALLY_SUPPORTED_CHECKSUMS.
        """
        self.logger(logging.DEBUG, 'rclone.stat: path: {}'.format(path))
        ret = {}
        chsum = None
        if path.startswith('rclone://'):
            path = self.pfn2path(path)

        try:
            # rclone stat for getting filesize
            cmd = 'rclone size {0}:{1}'.format(self.hostname, path)
            self.logger(logging.DEBUG, 'rclone.stat: filesize cmd: {}'.format(cmd))
            status_stat, out, err = execute(cmd)
            if status_stat == 0:
                fsize = (out.split('\n')[1]).split(' ')[4][1:]
                ret['filesize'] = fsize

            # rclone query checksum for getting md5 checksum
            cmd = 'rclone md5sum %s:%s' % (self.hostname, path)
            self.logger(logging.DEBUG, 'rclone.stat: checksum cmd: {}'.format(cmd))
            status_query, out, err = execute(cmd)

            if status_query == 0:
                chsum = 'md5'
                val = out.strip(' ').split()
                ret[chsum] = val[0]

        except Exception as e:
            raise exception.ServiceUnavailable(e)

        if 'filesize' not in ret:
            raise exception.ServiceUnavailable('Filesize could not be retrieved.')
        if PREFERRED_CHECKSUM != chsum or not chsum:
            msg = '{} does not match with {}'.format(chsum, PREFERRED_CHECKSUM)
            raise exception.RSEChecksumUnavailable(msg)

        return ret

    def pfn2path(self, pfn):
        """
        Returns the path of a file given the pfn, i.e. scheme, user and hostname are subtracted from the pfn.

        :param path: pfn of a file

        :returns: path.
        """
        path = pfn
        if pfn.startswith('rclone://'):
            self.logger(logging.DEBUG, 'rclone.pfn2path: pfn: {}'.format(pfn))
            prefix = self.attributes['prefix']
            path = pfn.partition(self.attributes['prefix'])[2]
            path = prefix + path
        return path

    def lfns2pfns(self, lfns):
        """
        Returns a fully qualified PFN for the file referred by path.

        :param path: The path to the file.

        :returns: Fully qualified PFN.
        """
        self.logger(logging.DEBUG, 'rclone.lfns2pfns: lfns: {}'.format(lfns))
        pfns = {}
        prefix = self.attributes['prefix']

        if not prefix.startswith('/'):
            prefix = ''.join(['/', prefix])
        if not prefix.endswith('/'):
            prefix = ''.join([prefix, '/'])

        lfns = [lfns] if isinstance(lfns, dict) else lfns
        for lfn in lfns:
            scope, name = lfn['scope'], lfn['name']
            if 'path' in lfn and lfn['path'] is not None:
                pfns['%s:%s' % (scope, name)] = ''.join([self.attributes['scheme'], '://', self.host, ':', prefix, lfn['path']])
            else:
                pfns['%s:%s' % (scope, name)] = ''.join([self.attributes['scheme'], '://', self.host, ':', prefix, self._get_path(scope=scope, name=name)])
        return pfns

    def connect(self):
        """ Establishes the actual connection to the referred RSE.

            :raises RSEAccessDenied
        """
        self.logger(logging.DEBUG, 'rclone.connect: hostname {}'.format(self.hostname))
        try:
            cmd = 'rclone lsd %s:' % (self.hostname)
            status, out, err = execute(cmd)
            if status:
                raise exception.RSEAccessDenied(err)
        except Exception as e:
            raise exception.RSEAccessDenied(e)

    def close(self):
        """ Closes the connection to RSE."""
        pass

    def get(self, pfn, dest, transfer_timeout=None):
        """ Provides access to files stored inside connected the RSE.

            :param pfn: Physical file name of requested file
            :param dest: Name and path of the files when stored at the client
            :param transfer_timeout: Transfer timeout (in seconds) - dummy

            :raises DestinationNotAccessible, ServiceUnavailable, SourceNotFound
        """
        self.logger(logging.DEBUG, 'rclone.get: pfn: {}'.format(pfn))
        try:
            path = self.pfn2path(pfn)
            cmd = 'rclone copyto %s:%s %s' % (self.hostname, path, dest)
            self.logger(logging.DEBUG, 'rclone.get: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def put(self, filename, target, source_dir, transfer_timeout=None):
        """
            Allows to store files inside the referred RSE.

            :param source: path to the source file on the client file system
            :param target: path to the destination file on the storage
            :param source_dir: Path where the to be transferred files are stored in the local file system
            :param transfer_timeout: Transfer timeout (in seconds) - dummy

            :raises DestinationNotAccessible: if the destination storage was not accessible.
            :raises ServiceUnavailable: if some generic error occurred in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        self.logger(logging.DEBUG, 'rclone.put: filename: {} target: {}'.format(filename, target))
        source_dir = source_dir or '.'
        source_url = '%s/%s' % (source_dir, filename)
        self.logger(logging.DEBUG, 'rclone.put: source url: {}'.format(source_url))

        path = self.pfn2path(target)
        if not os.path.exists(source_url):
            raise exception.SourceNotFound()
        try:
            cmd = 'rclone copyto %s %s:%s' % (source_url, self.hostname, path)
            self.logger(logging.DEBUG, 'rclone.put: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def delete(self, pfn):
        """
            Deletes a file from the connected RSE.

            :param pfn: Physical file name

            :raises ServiceUnavailable: if some generic error occurred in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        self.logger(logging.DEBUG, 'rclone.delete: pfn: {}'.format(pfn))
        if not self.exists(pfn):
            raise exception.SourceNotFound()
        try:
            path = self.pfn2path(pfn)
            cmd = 'rclone delete %s:%s' % (self.hostname, path)
            self.logger(logging.DEBUG, 'rclone.delete: cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status != 0:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)

    def rename(self, pfn, new_pfn):
        """ Allows to rename a file stored inside the connected RSE.

            :param pfn:      Current physical file name
            :param new_pfn  New physical file name
            :raises DestinationNotAccessible: if the destination storage was not accessible.
            :raises ServiceUnavailable: if some generic error occurred in the library.
            :raises SourceNotFound: if the source file was not found on the referred storage.
        """
        self.logger(logging.DEBUG, 'rclone.rename: pfn: {}'.format(pfn))
        if not self.exists(pfn):
            raise exception.SourceNotFound()
        try:
            path = self.pfn2path(pfn)
            new_path = self.pfn2path(new_pfn)
            cmd = 'rclone moveto %s:%s %s:%s' % (self.hostname, path, self.hostname, new_path)
            self.logger(logging.DEBUG, 'rclone.stat: rename cmd: {}'.format(cmd))
            status, out, err = execute(cmd)
            if status != 0:
                raise exception.RucioException(err)
        except Exception as e:
            raise exception.ServiceUnavailable(e)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import click

from rucio.cli.bin_legacy.rucio import upload
from rucio.cli.utils import Arguments
from rucio.common.config import config_get_float


@click.command("upload")
@click.argument("file-paths", nargs=-1)
@click.option("--rse", "--rse-name", help="Rucio Storage Element (RSE) name", required=True)
@click.option("--lifetime", type=int, help="Lifetime of the rule in seconds")
@click.option("--expiration-date", help="The date when the rule expires in UTC, format: <year>-<month>-<day>-<hour>:<minute>:<second>. E.g. 2022-10-20-20:00:00")
@click.option("--scope", help="Scope name.")
@click.option("--impl", type=click.Choice([]), help="Transfer protocol implementation to use (e.g: xrootd, gfal.NoRename, webdav, ssh.Rsync, rclone)")
# The --no-register option is hidden. This is pilot ONLY. Users should not use this. Will lead to unregistered data on storage!
@click.option("--no-register", is_flag=True, default=False, hidden=True)
@click.option("--register-after-upload", is_flag=True, default=False, help="Register the file _only_ after successful upload")
@click.option("--summary", is_flag=True, default=False, help="Create rucio_upload.json summary file")
@click.option("--guid", help="Manually specify the GUID for the file.")
@click.option("--protocol", help="Force the protocol to use")
@click.option("--pfn", help="Specify the exact PFN for the upload")
@click.option("--lfn", help="Specify the exact LFN for the upload")
@click.option("--transfer-timeout", type=float, default=config_get_float("upload", "transfer_timeout", False, 360), help="Transfer timeout (in seconds)")
@click.option("-r", "--recursive", is_flag=True, default=False, help="Convert recursively the folder structure into collections")
@click.pass_context
def upload_command(ctx, file_paths, rse, lifetime, expiration_date, scope, impl, no_register, register_after_upload, summary, guid, protocol, pfn, lfn, transfer_timeout, recursive):
    """Upload file(s) to a Rucio RSE"""
    args = Arguments(
        {
            "args": file_paths,
            "rse": rse,
            "lifetime": lifetime,
            "expiration_date": expiration_date,
            "scope": scope,
            "impl": impl,
            "no_register": no_register,
            "register_after_upload": register_after_upload,
            "protocol": protocol,
            "summary": summary,
            "guid": guid,
            "pfn": pfn,
            "name": lfn,
            "transfer_timeout": transfer_timeout,
            "recursive": recursive,
        }
    )
    upload(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from logging import getLogger
from os import remove
from random import choice
from re import search
from string import ascii_letters, ascii_lowercase, ascii_uppercase, digits
from unittest.mock import patch
from urllib.parse import parse_qs, urlparse

import pytest
from oic import rndstr
from sqlalchemy import select

from rucio.client.accountlimitclient import AccountLimitClient
from rucio.client.client import Client
from rucio.client.replicaclient import ReplicaClient
from rucio.client.subscriptionclient import SubscriptionClient
from rucio.client.uploadclient import UploadClient
from rucio.common.config import config_add_section, config_has_section, config_remove_option, config_set
from rucio.common.constants import RseAttr
from rucio.common.exception import AccessDenied, Duplicate, InvalidRSEExpression, RucioException, UnsupportedAccountName, UnsupportedOperation
from rucio.common.types import InternalAccount
from rucio.common.utils import generate_uuid, get_tmp_dir, parse_response, ssh_sign
from rucio.core import config as core_config
from rucio.core.account_counter import add_counter
from rucio.core.replica import add_replica
from rucio.core.rse import get_rse_id, get_rse_vo, get_rses_with_attribute_value
from rucio.core.rse_expression_parser import parse_expression
from rucio.core.rule import add_rule
from rucio.core.vo import map_vo
from rucio.daemons.automatix.automatix import automatix
from rucio.db.sqla import models
from rucio.db.sqla import session as db_session
from rucio.gateway import vo as vo_gateway
from rucio.gateway.account import add_account, list_accounts
from rucio.gateway.account_limit import set_local_account_limit
from rucio.gateway.authentication import get_auth_token_gss, get_auth_token_saml, get_auth_token_x509
from rucio.gateway.did import add_did, list_dids
from rucio.gateway.identity import add_account_identity, list_accounts_for_identity
from rucio.gateway.lock import get_replica_locks_for_rule_id
from rucio.gateway.replica import list_replicas
from rucio.gateway.rse import add_protocol, add_rse, add_rse_attribute, list_rses
from rucio.gateway.rule import delete_replication_rule, get_replication_rule
from rucio.gateway.scope import add_scope, list_scopes
from rucio.gateway.subscription import add_subscription, list_subscriptions
from rucio.tests.common import auth, execute, hdrdict, headers, loginhdr, vohdr

from .test_authentication import PRIVATE_KEY, PUBLIC_KEY
from .test_oidc import NEW_TOKEN_DICT, get_mock_oidc_client

LOG = getLogger(__name__)

# module-level skip, see https://docs.pytest.org/en/latest/skipping.html#skip-all-test-functions-of-a-class-or-module
pytestmark = pytest.mark.skipif('SUITE' in os.environ and os.environ['SUITE'] != 'multi_vo',
                                reason='No execution of the multi_vo tests in a suite other than the multi_vo suite')


class TestVOCoreGateway:

    @pytest.mark.noparallel(reason='changes global configuration value')
    def test_multi_vo_flag(self, vo, second_vo):
        """ MULTI VO (CORE): Test operations fail in single_vo mode """
        try:
            config_set('common', 'multi_vo', 'False')
            with pytest.raises(UnsupportedOperation):
                vo_gateway.list_vos(issuer='super_root', vo='def')
            config_remove_option('common', 'multi_vo')
            with pytest.raises(UnsupportedOperation):
                vo_gateway.list_vos(issuer='super_root', vo='def')
        finally:
            config_set('common', 'multi_vo', 'True')

    @pytest.mark.noparallel(reason='uses global RSE (MOCK) and fails when run in parallel')
    def test_access_rule_vo(self, vo, second_vo, root_account, mock_scope):
        """ MULTI VO (CORE): Test accessing rules from a different VO """
        scope = mock_scope
        dataset = 'dataset_' + str(generate_uuid())
        rse_str = ''.join(choice(ascii_uppercase) for x in range(10))
        rse_name = 'MOCK_%s' % rse_str
        rse_id = add_rse(rse_name, 'root', vo=vo)

        add_replica(rse_id=rse_id, scope=scope, name=dataset, bytes_=10, account=root_account)
        rule_id = add_rule(dids=[{'scope': scope, 'name': dataset}], account=root_account, copies=1, rse_expression='MOCK', grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        with pytest.raises(AccessDenied):
            delete_replication_rule(rule_id=rule_id, purge_replicas=False, issuer='root', vo=second_vo)

        # check locks are not accessible from other VO
        locks = list(get_replica_locks_for_rule_id(rule_id, vo=vo))
        assert len(locks) == 1
        locks = list(get_replica_locks_for_rule_id(rule_id, vo=second_vo))
        assert len(locks) == 0

        delete_replication_rule(rule_id=rule_id, purge_replicas=False, issuer='root', vo=vo)
        rule_dict = get_replication_rule(rule_id=rule_id, issuer='root', vo=vo)
        assert rule_dict['expires_at'] is not None

    def test_add_vo(self, vo, second_vo):
        """ MULTI VO (CORE): Test creation of VOs """
        with pytest.raises(AccessDenied):
            vo_gateway.add_vo(second_vo, 'root', 'Add new VO with root', 'rucio@email.com', vo=vo)
        with pytest.raises(Duplicate):
            vo_gateway.add_vo(second_vo, 'super_root', 'Add existing VO', 'rucio@email.com', 'def')

    def test_recover_root_identity(self, vo, second_vo):
        """ MULTI VO (CORE): Test adding a new identity for root using super_root """
        identity_key = ''.join(choice(ascii_lowercase) for x in range(10))
        with pytest.raises(AccessDenied):
            vo_gateway.recover_vo_root_identity(root_vo=second_vo, identity_key=identity_key, id_type='userpass', email='rucio@email.com', issuer='root', password='password', vo=vo)
        vo_gateway.recover_vo_root_identity(root_vo=second_vo, identity_key=identity_key, id_type='userpass', email='rucio@email.com', issuer='super_root', password='password', vo='def')
        assert 'root' in list_accounts_for_identity(identity_key=identity_key, id_type='userpass')

    def test_update_vo(self, vo, second_vo):
        """ MULTI VO (CORE): Test updating VOs """
        description = generate_uuid()
        email = generate_uuid()
        parameters = {'vo': second_vo, 'description': description, 'email': email}
        with pytest.raises(AccessDenied):
            vo_gateway.update_vo(second_vo, parameters, 'root', vo=vo)
        vo_gateway.update_vo(second_vo, parameters, 'super_root', 'def')
        vo_update_success = False
        for v in vo_gateway.list_vos('super_root', 'def'):
            if v['vo'] == parameters['vo']:
                assert email == v['email']
                assert description == v['description']
                vo_update_success = True
        assert vo_update_success

    def test_super_root_permissions(self, vo, second_vo):
        """ MULTI VO (CORE): Test super_root cannot access root/user functions """
        rse_str = ''.join(choice(ascii_uppercase) for x in range(10))
        rse_name = 'MOCK_%s' % rse_str
        scope_uuid = str(generate_uuid()).lower()[:16]
        scope = 'mock_%s' % scope_uuid

        # Test super_root@def with functions at vo='def'
        with pytest.raises(AccessDenied):
            add_rse(rse_name, 'super_root', vo='def')
        with pytest.raises(AccessDenied):
            add_scope(scope, 'super_root', 'super_root', vo='def')

    @pytest.mark.noparallel(reason='changes global configuration value')
    def test_super_root_naming(self, vo, second_vo):
        """ MULTI VO (CORE): Test we can only name accounts super_root when appropriate """
        with pytest.raises(Duplicate):  # Ensure we fail from duplication rather than the choice of name
            add_account('super_root', 'USER', 'rucio@email.com', 'root', vo='def')
        with pytest.raises(UnsupportedAccountName):
            add_account('super_root', 'USER', 'rucio@email.com', 'root', vo=vo)
        try:
            config_remove_option('common', 'multi_vo')
            with pytest.raises(UnsupportedAccountName):
                add_account('super_root', 'USER', 'rucio@email.com', 'root', vo=vo)
            with pytest.raises(UnsupportedAccountName):
                add_account('super_root', 'USER', 'rucio@email.com', 'root', vo='def')
        finally:
            # Make sure we don't leave the config changed due to a test failure
            config_set('common', 'multi_vo', 'True')


@pytest.fixture(scope='class')
def usr_uuid():
    return str(generate_uuid()).lower()[:16]


@pytest.fixture(scope='class')
def account_tst(usr_uuid, vo):
    account_tst = 'tst-%s' % usr_uuid
    add_account(account_tst, 'USER', 'rucio@email.com', 'root', vo=vo)
    yield account_tst


@pytest.fixture(scope='class')
def account_new(usr_uuid, second_vo):
    account_new = 'new-%s' % usr_uuid
    add_account(account_new, 'USER', 'rucio@email.com', 'root', vo=second_vo)
    yield account_new


class TestVORestAPI:

    @staticmethod
    def auth_oidc_handling(mock_oidc_client, rest_client, vo, long_vo, account_in, account_not_in, auto, polling):
        """
        Utility script to handle the REST calls with various urls and codes needed to authenticate via OIDC.
        IdP responses are faked using code from `test_oidc.py`.

        :param mock_oidc_client: Mock OIDC client used to fake responses from the IdP for test purposes.
        :param vo: VO to authenticate
        :param account_in: A string (externally) representing an account we DO expect to find at the VO.
        :param account_not_in: A string (externally) representing an account we DO NOT expect to find at the VO.
        :param auto: Boolean to specify whether we automatically submit userpass to the IdP as part of authentication.
        :param auto: Boolean to specify whether we poll the IdP for a successful login as part of authentication.
        """
        mock_oidc_client.side_effect = get_mock_oidc_client

        try:
            add_account_identity('SUB=knownsub, ISS=https://test_issuer/', 'OIDC', 'root', 'rucio_test@test.com', 'root', vo=vo)
        except Duplicate:
            pass  # Might already exist, can skip

        # Define headers
        headers_dict = {'X-Rucio-Account': 'root',
                        'X-Rucio-VO': long_vo,
                        'X-Rucio-Client-Authorize-Auto': str(auto),
                        'X-Rucio-Client-Authorize-Polling': str(polling),
                        'X-Rucio-Client-Authorize-Scope': 'openid profile',
                        'X-Rucio-Client-Authorize-Refresh-Lifetime': '96',
                        'X-Rucio-Client-Authorize-Audience': 'rucio',
                        'X-Rucio-Client-Authorize-Issuer': 'dummy_admin_iss_nickname'}

        response = rest_client.get('/auth/oidc', headers=headers(hdrdict(headers_dict)))
        assert response.status_code == 200
        if auto:
            # Get the auth_url without any redirect
            auth_url = response.headers.get('X-Rucio-OIDC-Auth-URL')
        else:
            # Get the redirect_url
            redirect_url = response.headers.get('X-Rucio-OIDC-Auth-URL')
            assert 'https://test_redirect_string/auth/oidc_redirect?' in redirect_url
            if polling:
                assert '_polling' in redirect_url
            else:
                assert '_polling' not in redirect_url
            redirect_url_parsed = urlparse(redirect_url)

            # Get the auth_url from the redirect_url
            response = rest_client.get('/auth/oidc_redirect?%s' % redirect_url_parsed.query, headers=headers(hdrdict(headers_dict)))
            assert response.status_code == 303
            auth_url = response.headers.get('location')

        assert 'https://test_auth_url_string?' in auth_url
        auth_url_parsed = urlparse(auth_url)
        auth_url_params = parse_qs(auth_url_parsed.query)

        # Fake the IdP response for a successful login
        code_response = rndstr()
        access_token = rndstr()
        NEW_TOKEN_DICT['access_token'] = access_token
        NEW_TOKEN_DICT['id_token'] = {'sub': 'knownsub', 'iss': 'https://test_issuer/', 'nonce': auth_url_params['nonce'][0]}
        headers_dict['X-Rucio-Client-Fetch-Token'] = 'True'

        if auto:
            # Can get the token now
            response = rest_client.get('/auth/oidc_token?state=%s&code=%s' % (auth_url_params['state'][0], code_response), headers=headers(hdrdict(headers_dict)))
        else:
            # Get the html response
            response = rest_client.get('/auth/oidc_code?state=%s&code=%s' % (auth_url_params['state'][0], code_response), headers=headers(hdrdict(headers_dict)))
            assert response.status_code == 200
            if polling:
                assert 'Rucio Client should now be able to fetch your token automatically.' in response.get_data(as_text=True)
                response = rest_client.get('/auth/oidc_redirect?%s' % redirect_url_parsed.query, headers=headers(hdrdict(headers_dict)))
            else:
                # Get the fetch_code from the response, then submit it
                fetch_code = search(r'<b>[a-zA-Z0-9]{50}</b>', response.get_data(as_text=True))
                assert fetch_code is not None
                fetch_code = fetch_code.group()[3:53]
                response = rest_client.get('/auth/oidc_redirect?%s' % fetch_code, headers=headers(hdrdict(headers_dict)))

        # Regardless of how we got it, check we have the token and that we only get results from our VO when using it
        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))
        response = rest_client.get('/accounts/', headers=headers(auth(token)))
        assert response.status_code == 200
        accounts = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts) != 0
        assert account_in in accounts
        assert account_not_in not in accounts

    @patch('rucio.core.oidc.__get_init_oidc_client')
    def test_auth_oidc(self, mock_oidc_client, vo, long_vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test oidc authentication to multiple VOs """
        self.auth_oidc_handling(mock_oidc_client, rest_client, vo, long_vo, account_tst, account_new, auto=False, polling=False)
        self.auth_oidc_handling(mock_oidc_client, rest_client, second_vo, second_vo, account_new, account_tst, auto=False, polling=False)

    @patch('rucio.core.oidc.__get_init_oidc_client')
    def test_auth_oidc_polling(self, mock_oidc_client, vo, long_vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test oidc authentication to multiple VOs using 'polling' option """
        self.auth_oidc_handling(mock_oidc_client, rest_client, vo, long_vo, account_tst, account_new, auto=False, polling=True)
        self.auth_oidc_handling(mock_oidc_client, rest_client, second_vo, second_vo, account_new, account_tst, auto=False, polling=True)

    @patch('rucio.core.oidc.__get_init_oidc_client')
    def test_auth_oidc_auto(self, mock_oidc_client, vo, long_vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test oidc authentication to multiple VOs using 'auto' option """
        self.auth_oidc_handling(mock_oidc_client, rest_client, vo, long_vo, account_tst, account_new, auto=True, polling=False)
        self.auth_oidc_handling(mock_oidc_client, rest_client, second_vo, second_vo, account_new, account_tst, auto=True, polling=False)

    def test_auth_gss(self, vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test gss authentication to multiple VOs """
        # Can't rely on `requests_kerberos` module being present, so get tokens from gateway instead
        token_tst = get_auth_token_gss('root', 'rucio-dev@CERN.CH', 'unknown', None, vo=vo).get('token')
        token_new = get_auth_token_gss('root', 'rucio-dev@CERN.CH', 'unknown', None, vo=second_vo).get('token')

        response = rest_client.get('/accounts/', headers=headers(auth(token_tst)))
        assert response.status_code == 200
        accounts_tst = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_tst) != 0
        assert account_tst in accounts_tst
        assert account_new not in accounts_tst

        response = rest_client.get('/accounts/', headers=headers(auth(token_new)))
        assert response.status_code == 200
        accounts_new = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_new) != 0
        assert account_new in accounts_new
        assert account_tst not in accounts_new

    def test_auth_saml(self, vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test saml authentication to multiple VOs """
        try:
            add_account_identity('ddmlab', 'SAML', 'root', 'rucio@email.com', 'root', vo=vo)
        except Duplicate:
            pass  # Might already exist, can skip

        try:
            add_account_identity('ddmlab', 'SAML', 'root', 'rucio@email.com', 'root', vo=second_vo)
        except Duplicate:
            pass  # Might already exist, can skip

        # Can't rely on `onelogin` module being present, so get tokens from gateway instead
        token_tst = get_auth_token_saml('root', 'ddmlab', 'unknown', None, vo=vo).get('token')
        token_new = get_auth_token_saml('root', 'ddmlab', 'unknown', None, vo=second_vo).get('token')

        response = rest_client.get('/accounts/', headers=headers(auth(token_tst)))
        assert response.status_code == 200
        accounts_tst = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_tst) != 0
        assert account_tst in accounts_tst
        assert account_new not in accounts_tst

        response = rest_client.get('/accounts/', headers=headers(auth(token_new)))
        assert response.status_code == 200
        accounts_new = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_new) != 0
        assert account_new in accounts_new
        assert account_tst not in accounts_new

    def test_auth_ssh(self, vo, long_vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test ssh authentication to multiple VOs """
        try:
            add_account_identity(PUBLIC_KEY, 'SSH', 'root', 'rucio@email.com', 'root', vo=vo)
        except Duplicate:
            pass  # Might already exist, can skip

        try:
            add_account_identity(PUBLIC_KEY, 'SSH', 'root', 'rucio@email.com', 'root', vo=second_vo)
        except Duplicate:
            pass  # Might already exist, can skip

        headers_dict = {'X-Rucio-Account': 'root'}
        response = rest_client.get('/auth/ssh_challenge_token', headers=headers(hdrdict(headers_dict), vohdr(long_vo)))
        assert response.status_code == 200
        challenge_tst = str(response.headers.get('X-Rucio-SSH-Challenge-Token'))
        headers_dict.update({'X-Rucio-SSH-Signature': ssh_sign(PRIVATE_KEY, challenge_tst)})
        response = rest_client.get('/auth/ssh', headers=headers(hdrdict(headers_dict), vohdr(long_vo)))
        assert response.status_code == 200
        token_tst = str(response.headers.get('X-Rucio-Auth-Token'))

        headers_dict = {'X-Rucio-Account': 'root'}
        response = rest_client.get('/auth/ssh_challenge_token', headers=headers(hdrdict(headers_dict), vohdr(second_vo)))
        assert response.status_code == 200
        challenge_tst = str(response.headers.get('X-Rucio-SSH-Challenge-Token'))
        headers_dict.update({'X-Rucio-SSH-Signature': ssh_sign(PRIVATE_KEY, challenge_tst)})
        response = rest_client.get('/auth/ssh', headers=headers(hdrdict(headers_dict), vohdr(second_vo)))
        assert response.status_code == 200
        token_new = str(response.headers.get('X-Rucio-Auth-Token'))

        response = rest_client.get('/accounts/', headers=headers(auth(token_tst)))
        assert response.status_code == 200
        accounts_tst = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_tst) != 0
        assert account_tst in accounts_tst
        assert account_new not in accounts_tst

        response = rest_client.get('/accounts/', headers=headers(auth(token_new)))
        assert response.status_code == 200
        accounts_new = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_new) != 0
        assert account_new in accounts_new
        assert account_tst not in accounts_new

    def test_auth_userpass(self, vo, long_vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test userpass authentication to multiple VOs """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr(long_vo)))
        assert response.status_code == 200
        token_tst = str(response.headers.get('X-Rucio-Auth-Token'))

        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr(second_vo)))
        assert response.status_code == 200
        token_new = str(response.headers.get('X-Rucio-Auth-Token'))

        response = rest_client.get('/accounts/', headers=headers(auth(token_tst)))
        assert response.status_code == 200
        accounts_tst = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_tst) != 0
        assert account_tst in accounts_tst
        assert account_new not in accounts_tst

        response = rest_client.get('/accounts/', headers=headers(auth(token_new)))
        assert response.status_code == 200
        accounts_new = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_new) != 0
        assert account_new in accounts_new
        assert account_tst not in accounts_new

    def test_auth_x509(self, vo, second_vo, account_tst, account_new, rest_client):
        """ MULTI VO (REST): Test X509 authentication to multiple VOs """
        # Flasks test client doesn't support client certificates, so get tokens from gateway instead
        token_tst = get_auth_token_x509('root', 'CN=Rucio User', 'unknown', None, vo=vo).get('token')
        token_new = get_auth_token_x509('root', 'CN=Rucio User', 'unknown', None, vo=second_vo).get('token')

        response = rest_client.get('/accounts/', headers=headers(auth(token_tst)))
        assert response.status_code == 200
        accounts_tst = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_tst) != 0
        assert account_tst in accounts_tst
        assert account_new not in accounts_tst

        response = rest_client.get('/accounts/', headers=headers(auth(token_new)))
        assert response.status_code == 200
        accounts_new = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(accounts_new) != 0
        assert account_new in accounts_new
        assert account_tst not in accounts_new

    def test_list_vos_success(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test list VOs through REST layer succeeds """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        response = rest_client.get('/vos/', headers=headers(auth(token)))
        assert response.status_code == 200
        vo_dicts = [parse_response(r) for r in response.get_data(as_text=True).split('\n')[:-1]]
        assert len(vo_dicts) != 0
        for vo_dict in vo_dicts:
            assert vo_dict['vo'] is not None
            assert vo_dict['email'] is not None
            assert vo_dict['description'] is not None
            assert vo_dict['created_at'] is not None
            assert vo_dict['updated_at'] is not None

    def test_list_vos_denied(self, vo, long_vo, second_vo, rest_client):
        """ MULTI VO (REST): Test list VOs through REST layer raises AccessDenied """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr(long_vo)))
        assert response.status_code == 200
        token = response.headers.get('X-Rucio-Auth-Token')
        assert token

        response = rest_client.get('/vos/', headers=headers(auth(token)))
        assert response.status_code == 401

    @pytest.mark.noparallel(reason='changes global configuration value')
    def test_list_vos_unsupported(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test list VOs through REST layer raises UnsupportedOperation """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        try:
            config_set('common', 'multi_vo', 'False')
            response = rest_client.get('/vos/', headers=headers(auth(token)))
            assert response.status_code == 409

            config_remove_option('common', 'multi_vo')
            response = rest_client.get('/vos/', headers=headers(auth(token)))
            assert response.status_code == 409

        finally:
            # Make sure we don't leave the config changed due to a test failure
            config_set('common', 'multi_vo', 'True')

    def test_add_vo_denied(self, vo, long_vo, second_vo, rest_client):
        """ MULTI VO (REST): Test adding VO through REST layer raises AccessDenied """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr(long_vo)))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        params = {'email': 'rucio@email.com', 'decription': 'Try adding with root'}
        response = rest_client.post('/vos/' + vo, headers=headers(auth(token)), json=params)
        assert response.status_code == 401

    @pytest.mark.noparallel(reason='changes global configuration value')
    def test_add_vo_unsupported(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test adding VO through REST layer raises UnsupportedOperation """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        params = {'email': 'rucio@email.com', 'decription': 'Try adding in single vo mode'}
        try:
            config_set('common', 'multi_vo', 'False')
            response = rest_client.post('/vos/' + vo, headers=headers(auth(token)), json=params)
            assert response.status_code == 409

            config_remove_option('common', 'multi_vo')
            response = rest_client.post('/vos/' + vo, headers=headers(auth(token)), json=params)
            assert response.status_code == 409

        finally:
            # Make sure we don't leave the config changed due to a test failure
            config_set('common', 'multi_vo', 'True')

    def test_add_vo_duplicate(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test adding VO through REST layer raises Duplicate """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        params = {'email': 'rucio@email.com', 'decription': 'Try adding duplicate'}
        response = rest_client.post('/vos/' + vo, headers=headers(auth(token)), json=params)
        assert response.status_code == 409

    def test_update_vo_success(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test updating VO through REST layer succeeds """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        params = {'email': generate_uuid(), 'description': generate_uuid()}
        response = rest_client.put('/vos/' + vo, headers=headers(auth(token)), json=params)
        assert response.status_code == 200

        vo_update_success = False
        for v in vo_gateway.list_vos('super_root', 'def'):
            if v['vo'] == vo:
                assert params['email'] == v['email']
                assert params['description'] == v['description']
                vo_update_success = True
        assert vo_update_success

    def test_update_vo_denied(self, vo, long_vo, second_vo, rest_client):
        """ MULTI VO (REST): Test updating VO through REST layer raises AccessDenied """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr(long_vo)))
        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        params = {'email': 'rucio@email.com', 'decription': 'Try updating with root'}
        response = rest_client.put('/vos/' + vo, headers=headers(auth(token)), json=params)
        assert response.status_code == 401

    @pytest.mark.noparallel(reason='changes global configuration value')
    def test_update_vo_unsupported(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test updating VO through REST layer raises UnsupportedOperation """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))
        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        params = {'email': 'rucio@email.com', 'decription': 'Try updating in single vo mode'}
        try:
            config_set('common', 'multi_vo', 'False')
            response = rest_client.put('/vos/' + vo, headers=headers(auth(token)), json=params)
            assert response.status_code == 409

            config_remove_option('common', 'multi_vo')
            response = rest_client.put('/vos/' + vo, headers=headers(auth(token)), json=params)
            assert response.status_code == 409

        finally:
            # Make sure we don't leave the config changed due to a test failure
            config_set('common', 'multi_vo', 'True')

    def test_update_vo_not_found(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test updating VO through REST layer raises VONotFound """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        params = {'email': 'rucio@email.com', 'decription': 'Try updating non-existent'}
        response = rest_client.put('/vos/bad', headers=headers(auth(token)), json=params)
        assert response.status_code == 404

    def test_recover_vo_success(self, vo, second_vo, rest_client):
        """ MULTI VO (REST): Test recovering VO through REST layer succeeds """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('super_root', 'ddmlab', 'secret'), vohdr('def')))

        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        identity_key = ''.join(choice(ascii_lowercase) for x in range(10))
        params = {'identity': identity_key, 'authtype': 'userpass', 'email': 'rucio@email.com', 'password': 'password'}
        response = rest_client.post('/vos/' + vo + '/recover', headers=headers(auth(token)), json=params)
        assert response.status_code == 201

        assert 'root' in list_accounts_for_identity(identity_key=identity_key, id_type='userpass')

    def test_recover_vo_denied(self, vo, long_vo, rest_client):
        """ MULTI VO (REST): Test recovering VO through REST layer raises AccessDenied """
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr(long_vo)))
        assert response.status_code == 200
        token = str(response.headers.get('X-Rucio-Auth-Token'))

        identity_key = ''.join(choice(ascii_lowercase) for x in range(10))
        params = {'identity': identity_key, 'authtype': 'userpass', 'email': 'rucio@email.com', 'password': 'password'}
        response = rest_client.post('/vos/' + vo + '/recover', headers=headers(auth(token)), json=params)
        assert response.status_code == 401

    @pytest.mark.noparallel(reason='account lists may be changed by other tests')
    def test_rest_vomap(self, vo, long_vo, rest_client):
        """ MULTI VO (REST): Test that both the long and short version of a VO name return the same results. """
        def get_vo_accounts(vo):
            response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr(vo)))
            assert response.status_code == 200
            token = str(response.headers.get('X-Rucio-Auth-Token'))
            response = rest_client.get('/accounts/', headers=headers(auth(token)))
            assert response.status_code == 200
            accounts = [parse_response(a)['account'] for a in response.get_data(as_text=True).split('\n')[:-1]]
            return sorted(accounts)
        # The test VOs contain different account names
        # We get all the account names with the long VO name and short VO name and check they are equal
        accounts_long = get_vo_accounts(long_vo)
        accounts_short = get_vo_accounts(vo)
        assert len(accounts_short) > 0
        assert accounts_short == accounts_long

    def test_rest_vomap_bad(self, rest_client):
        """ MULTI VO (REST): Test that we get a bad parameter (400) error with an invalid (out of spec) VO name. """
        # VO names cannot include an exclaimation mark
        response = rest_client.get('/auth/userpass', headers=headers(loginhdr('root', 'ddmlab', 'secret'), vohdr("BadVO!")))
        assert response.status_code == 400


class TestMultiVoClients:

    def test_get_vo_from_config(self, long_vo):
        """ MULTI VO (CLIENT): Get vo from config file when starting clients """
        # Start clients with vo explicitly set to None
        replica_client = ReplicaClient(vo=None)
        client = Client(vo=None)
        upload_client = UploadClient(_client=client)
        # Check the vo has been got from the config file
        assert replica_client.vo == long_vo
        assert upload_client.client.vo == long_vo

    def test_accounts_at_different_vos(self, second_vo, account_client):
        """ MULTI VO (CLIENT): Test that accounts from 2nd vo don't interfere """
        usr_uuid = str(generate_uuid()).lower()[:16]
        tst = 'tst-%s' % usr_uuid
        new = 'new-%s' % usr_uuid
        shr = 'shr-%s' % usr_uuid
        account_client.add_account(tst, 'USER', 'rucio@email.com')
        account_client.add_account(shr, 'USER', 'rucio@email.com')
        add_account(new, 'USER', 'rucio@email.com', 'root', vo=second_vo)
        add_account(shr, 'USER', 'rucio@email.com', 'root', vo=second_vo)
        account_list_tst = [a['account'] for a in account_client.list_accounts()]
        account_list_new = [a['account'] for a in list_accounts(filter_={}, vo=second_vo)]
        assert tst in account_list_tst
        assert new not in account_list_tst
        assert shr in account_list_tst
        assert tst not in account_list_new
        assert new in account_list_new
        assert shr in account_list_new

    def test_dids_at_different_vos(self, vo, second_vo, did_client):
        """ MULTI VO (CLIENT): Test that dids from 2nd vo don't interfere """
        scope_uuid = str(generate_uuid()).lower()[:16]
        scope = 'shr_%s' % scope_uuid
        add_scope(scope, 'root', 'root', vo=vo)
        add_scope(scope, 'root', 'root', vo=second_vo)
        did_uuid = str(generate_uuid()).lower()[:16]
        tst = 'tstset_%s' % did_uuid
        new = 'newset_%s' % did_uuid
        shr = 'shrset_%s' % did_uuid
        did_client.add_did(scope, tst, 'DATASET')
        did_client.add_did(scope, shr, 'DATASET')
        add_did(scope, new, 'DATASET', 'root', vo=second_vo)
        add_did(scope, shr, 'DATASET', 'root', vo=second_vo)
        did_list_tst = list(did_client.list_dids(scope, {}))
        did_list_new = list(list_dids(scope, {}, vo=second_vo))
        assert tst in did_list_tst
        assert new not in did_list_tst
        assert shr in did_list_tst
        assert tst not in did_list_new
        assert new in did_list_new
        assert shr in did_list_new

    def test_rses_at_different_vos(self, vo, second_vo, rse_client):
        """ MULTI VO (CLIENT): Test that RSEs from 2nd vo don't interfere """
        # Set up RSEs at two VOs
        rse_str = ''.join(choice(ascii_uppercase) for x in range(10))
        tst = 'TST_%s' % rse_str
        new = 'NEW_%s' % rse_str
        shr = 'SHR_%s' % rse_str
        rse_client.add_rse(tst)
        rse_client.add_rse(shr)
        add_rse(new, 'root', vo=second_vo)
        shr_id_new_original = add_rse(shr, 'root', vo=second_vo)  # Accurate rse_id for shared RSE at 'new'

        # Check the cached rse-id from each VO does not interfere
        shr_id_tst = get_rse_id(shr, vo=vo)
        shr_id_new = get_rse_id(shr, vo=second_vo)
        assert shr_id_new == shr_id_new_original
        assert shr_id_new != shr_id_tst

        # Check that when listing RSEs we only get RSEs for our VO
        rse_list_tst = [r['rse'] for r in rse_client.list_rses()]
        rse_list_new = [r['rse'] for r in list_rses(filters={}, vo=second_vo)]
        assert tst in rse_list_tst
        assert new not in rse_list_tst
        assert shr in rse_list_tst
        assert tst not in rse_list_new
        assert new in rse_list_new
        assert shr in rse_list_new

        # Check the cached attribute-value results do not interfere and only give results from the appropriate VO
        attribute_value = generate_uuid()
        add_rse_attribute(new, 'test', attribute_value, 'root', vo=second_vo)
        rses_tst_1 = list(get_rses_with_attribute_value('test', attribute_value, vo=vo))
        rses_new_1 = list(get_rses_with_attribute_value('test', attribute_value, vo=second_vo))
        rses_tst_2 = list(get_rses_with_attribute_value('test', attribute_value, vo=vo))
        rses_new_2 = list(get_rses_with_attribute_value('test', attribute_value, vo=second_vo))
        assert len(rses_tst_1) == 0
        assert len(rses_new_1) != 0
        assert len(rses_tst_2) == 0
        assert len(rses_new_2) != 0

        # check parse_expression
        rses_tst_3 = parse_expression(shr, filter_={'vo': vo})
        rses_tst_4 = parse_expression(tst, filter_={'vo': vo})
        rses_new_3 = parse_expression(shr, filter_={'vo': second_vo})
        with pytest.raises(InvalidRSEExpression):
            parse_expression(tst, filter_={'vo': second_vo})
        assert len(rses_tst_3) == 1
        assert shr_id_tst == rses_tst_3[0]['id']
        assert len(rses_tst_4) == 1
        assert tst == rses_tst_4[0]['rse']
        assert len(rses_new_3) == 1
        assert shr_id_new == rses_new_3[0]['id']

    def test_scopes_at_different_vos(self, second_vo, scope_client):
        """ MULTI VO (CLIENT): Test that scopes from 2nd vo don't interfere """
        scope_uuid = str(generate_uuid()).lower()[:16]
        tst = 'tst_%s' % scope_uuid
        new = 'new_%s' % scope_uuid
        shr = 'shr_%s' % scope_uuid
        scope_client.add_scope('root', tst)
        scope_client.add_scope('root', shr)
        add_scope(new, 'root', 'root', vo=second_vo)
        add_scope(shr, 'root', 'root', vo=second_vo)
        scope_list_tst = list(scope_client.list_scopes())
        scope_list_new = list(list_scopes(filter_={}, vo=second_vo))
        assert tst in scope_list_tst
        assert new not in scope_list_tst
        assert shr in scope_list_tst
        assert tst not in scope_list_new
        assert new in scope_list_new
        assert shr in scope_list_new

    def test_subscriptions_at_different_vos(self, second_vo, account_client, did_client, rse_client, scope_client):
        """ MULTI VO (CLIENT): Test that subscriptions from 2nd vo don't interfere """

        usr_uuid = str(generate_uuid()).lower()[:16]
        shr_acc = 'shr-%s' % usr_uuid
        account_client.add_account(shr_acc, 'USER', 'rucio@email.com')
        add_account(shr_acc, 'USER', 'rucio@email.com', 'root', vo=second_vo)

        scope_uuid = str(generate_uuid()).lower()[:16]
        tst_scope = 'tst_%s' % scope_uuid
        new_scope = 'new_%s' % scope_uuid
        scope_client.add_scope('root', tst_scope)
        add_scope(new_scope, 'root', 'root', vo=second_vo)

        did_uuid = str(generate_uuid()).lower()[:16]
        tst_did = 'tstset_%s' % did_uuid
        new_did = 'newset_%s' % did_uuid

        rse_str = ''.join(choice(ascii_uppercase) for x in range(10))
        tst_rse1 = 'TST1_%s' % rse_str
        tst_rse2 = 'TST2_%s' % rse_str
        new_rse1 = 'NEW1_%s' % rse_str
        new_rse2 = 'NEW2_%s' % rse_str
        rse_client.add_rse(tst_rse1)
        rse_client.add_rse(tst_rse2)
        add_rse(new_rse1, 'root', vo=second_vo)
        add_rse(new_rse2, 'root', vo=second_vo)

        acc_lim_client = AccountLimitClient()
        acc_lim_client.set_local_account_limit(shr_acc, tst_rse1, 10)
        acc_lim_client.set_local_account_limit(shr_acc, tst_rse2, 10)
        set_local_account_limit(shr_acc, new_rse1, 10, 'root', vo=second_vo)
        set_local_account_limit(shr_acc, new_rse2, 10, 'root', vo=second_vo)

        did_client.add_did(tst_scope, tst_did, 'DATASET', rse=tst_rse1)
        add_did(new_scope, new_did, 'DATASET', 'root', rse=new_rse1, vo=second_vo)

        sub_client = SubscriptionClient()
        sub_str = generate_uuid()
        tst_sub = 'tstsub_' + sub_str
        new_sub = 'newsub_' + sub_str
        shr_sub = 'shrsub_' + sub_str

        tst_sub_id = sub_client.add_subscription(tst_sub, shr_acc, {'scope': [tst_scope]},
                                                 [{'copies': 1, 'rse_expression': tst_rse2, 'weight': 0,
                                                   'activity': 'User Subscriptions'}],
                                                 '', None, 0, 0)
        shr_tst_sub_id = sub_client.add_subscription(shr_sub, shr_acc, {'scope': [tst_scope]},
                                                     [{'copies': 1, 'rse_expression': tst_rse2, 'weight': 0,
                                                       'activity': 'User Subscriptions'}],
                                                     '', None, 0, 0)

        new_sub_id = add_subscription(new_sub, shr_acc, {'scope': [new_scope]},
                                      [{'copies': 1, 'rse_expression': new_rse2, 'weight': 0, 'activity': 'User Subscriptions'}],
                                      '', False, 0, 0, 3, 'root', vo=second_vo)
        shr_new_sub_id = add_subscription(shr_sub, shr_acc, {'scope': [new_scope]},
                                          [{'copies': 1, 'rse_expression': new_rse2, 'weight': 0, 'activity': 'User Subscriptions'}],
                                          '', False, 0, 0, 3, 'root', vo=second_vo)

        tst_subs = [s['id'] for s in sub_client.list_subscriptions()]
        assert tst_sub_id in tst_subs
        assert shr_tst_sub_id in tst_subs
        assert new_sub_id not in tst_subs
        assert shr_new_sub_id not in tst_subs

        new_subs = [s['id'] for s in list_subscriptions(vo=second_vo)]
        assert new_sub_id in new_subs
        assert shr_new_sub_id in new_subs
        assert tst_sub_id not in new_subs
        assert shr_tst_sub_id not in new_subs

        shr_tst_subs = [s['id'] for s in sub_client.list_subscriptions(name=shr_sub)]
        assert shr_tst_sub_id in shr_tst_subs
        assert shr_new_sub_id not in shr_tst_subs

        shr_new_subs = [s['id'] for s in list_subscriptions(name=shr_sub, vo=second_vo)]
        assert shr_new_sub_id in shr_new_subs
        assert shr_tst_sub_id not in shr_new_subs

        acc_tst_subs = [s['id'] for s in sub_client.list_subscriptions(account=shr_acc)]
        assert tst_sub_id in acc_tst_subs
        assert shr_tst_sub_id in acc_tst_subs
        assert new_sub_id not in acc_tst_subs
        assert shr_new_sub_id not in acc_tst_subs

        acc_new_subs = [s['id'] for s in list_subscriptions(account=shr_acc, vo=second_vo)]
        assert new_sub_id in acc_new_subs
        assert shr_new_sub_id in acc_new_subs
        assert tst_sub_id not in acc_new_subs
        assert shr_tst_sub_id not in acc_new_subs

    def test_account_counters_at_different_vos(self, vo, second_vo, rse_client):
        """ MULTI VO (CLIENT): Test that account counters from 2nd vo don't interfere """

        session = db_session.get_session()

        # add some RSEs to test create_counters_for_new_account
        rse_str = ''.join(choice(ascii_uppercase) for x in range(10))
        tst_rse1 = 'TST1_%s' % rse_str
        new_rse1 = 'NEW1_%s' % rse_str
        rse_client.add_rse(tst_rse1)
        new_rse1_vo1_id = add_rse(new_rse1, 'root', vo=vo)
        new_rse1_vo2_id = add_rse(new_rse1, 'root', vo=second_vo)

        # add an account
        usr_uuid = str(generate_uuid()).lower()[:16]
        new_acc_str = 'shr-%s' % usr_uuid
        add_account(new_acc_str, 'USER', 'rucio@email.com', 'root', vo=vo)
        add_account(new_acc_str, 'USER', 'rucio@email.com', 'root', vo=second_vo)
        new_acc_vo1 = InternalAccount(new_acc_str, vo=vo)
        new_acc_vo2 = InternalAccount(new_acc_str, vo=second_vo)

        # Create counters
        add_counter(new_rse1_vo1_id, new_acc_vo1)
        add_counter(new_rse1_vo2_id, new_acc_vo2)

        stmt = select(
            models.AccountUsage.account,
            models.AccountUsage.rse_id
        ).distinct(
        ).where(
            models.AccountUsage.account == new_acc_vo2
        )
        acc_counters = session.execute(stmt).one()
        rse_id = acc_counters[1]
        vo = get_rse_vo(rse_id)
        assert vo == second_vo


class TestMultiVOBinRucio:

    marker = '$> '

    @pytest.fixture(autouse=True)
    def setup_object(self, vo, second_vo):
        # Setup RSEs at two VOs so we can determine which VO we authenticated against
        rse_str = ''.join(choice(ascii_uppercase) for x in range(10))
        self.rse_tst = 'TST_%s' % rse_str
        self.rse_new = 'NEW_%s' % rse_str
        add_rse(self.rse_tst, 'root', vo=vo)
        add_rse(self.rse_new, 'root', vo=second_vo)

        try:
            remove(get_tmp_dir() + '/.rucio_root@%s/auth_token_for_account_root' % vo)
        except OSError as e:
            if e.args[0] != 2:
                raise e
        try:
            remove(get_tmp_dir() + '/.rucio_root@%s/auth_token_for_account_root' % second_vo)
        except OSError as e:
            if e.args[0] != 2:
                raise e

    def test_vo_option_admin_cli(self, long_vo, second_vo):
        """ MULTI VO (USER): Test authentication to multiple VOs via the admin CLI """
        fake_vo = 'fke'

        cmd = 'rucio-admin --vo %s rse list' % long_vo
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, )
        assert self.rse_tst in out
        assert self.rse_new not in out

        cmd = 'rucio-admin --vo %s rse list' % second_vo
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, )
        assert self.rse_tst not in out
        assert self.rse_new in out

        cmd = 'rucio-admin --vo %s rse list' % fake_vo
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, err)
        assert 'Details: CannotAuthenticate' in err

        cmd = 'rucio-admin rse list'
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, )
        assert self.rse_tst in out
        assert self.rse_new not in out

    def test_vo_option_cli(self, long_vo, second_vo):
        """ MULTI VO (USER): Test authentication to multiple VOs via the CLI """
        fake_vo = 'fke'

        cmd = 'rucio --vo %s list-rses' % long_vo
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, )
        assert self.rse_tst in out
        assert self.rse_new not in out

        cmd = 'rucio --vo %s list-rses' % second_vo
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, )
        assert self.rse_tst not in out
        assert self.rse_new in out

        cmd = 'rucio --vo %s list-rses' % fake_vo
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, err)
        assert 'Details: CannotAuthenticate' in err

        cmd = 'rucio list-rses'
        print(self.marker + cmd)
        exitcode, out, err = execute(cmd)
        print(out, )
        assert self.rse_tst in out
        assert self.rse_new not in out


@pytest.mark.noparallel(reason='runs daemons, fails when run in parallel')
class TestMultiVODaemons:

    @pytest.mark.parametrize("core_config_mock", [{"table_content": []}], indirect=True)
    def test_automatix(self, second_vo, rse_client, did_client, replica_client, scope_client, core_config_mock):
        """ MULTI VO (DAEMON): Test that automatix runs on a single VO """
        scope_uuid = str(generate_uuid()).lower()[:16]
        shr_scope = 'shr_%s' % scope_uuid
        scope_client.add_scope('root', shr_scope)
        add_scope(shr_scope, 'root', 'root', vo=second_vo)

        rse_str = ''.join(choice(ascii_uppercase) for x in range(10))
        shr_rse = 'SHR_%s' % rse_str
        mock_protocol = {'scheme': 'MOCK',
                         'hostname': 'localhost',
                         'port': 123,
                         'prefix': '/test/automatix',
                         'impl': 'rucio.rse.protocols.mock.Default',
                         'domains': {
                             'lan': {'read': 1,
                                     'write': 1,
                                     'delete': 1},
                             'wan': {'read': 1,
                                     'write': 1,
                                     'delete': 1}}}
        rse_client.add_rse(shr_rse)
        rse_client.add_rse_attribute(rse=shr_rse, key=RseAttr.VERIFY_CHECKSUM, value=False)
        rse_client.add_rse_attribute(rse=shr_rse, key=RseAttr.SKIP_UPLOAD_STAT, value=True)
        rse_client.add_protocol(shr_rse, mock_protocol)
        add_rse(shr_rse, 'root', vo=second_vo)
        add_rse_attribute(rse=shr_rse, key=RseAttr.VERIFY_CHECKSUM, value=False, issuer='root', vo=second_vo)
        add_rse_attribute(rse=shr_rse, key=RseAttr.SKIP_UPLOAD_STAT, value=True, issuer='root', vo=second_vo)
        add_protocol(rse=shr_rse, data=mock_protocol, issuer='root', vo=second_vo)

        if not config_has_section("automatix"):
            config_add_section("automatix")
        config_set("automatix", "rses", shr_rse)
        config_set("automatix", "scope", shr_scope)

        automatix(
            inputfile='/opt/rucio/etc/automatix.json',
            sleep_time=10,
            once=True,
        )

        did_list_tst = list(did_client.list_dids(shr_scope, {}))
        did_list_new = list(list_dids(shr_scope, {}, vo=second_vo))
        assert len(did_list_tst) != 0
        assert len(did_list_new) == 0

        did_dicts = [{'scope': shr_scope, 'name': n} for n in did_list_tst]
        replicas_tst = list(replica_client.list_replicas(did_dicts, rse_expression=shr_rse))
        replicas_new = list(list_replicas(did_dicts, rse_expression=shr_rse, vo=second_vo))
        assert len(replicas_tst) != 0
        assert len(replicas_new) == 0


class TestVOMap:
    """ Test VO Mapping functions. """

    @pytest.mark.parametrize("core_config_mock", [{"table_content": []}], indirect=True)
    def test_map_vo(self, core_config_mock):
        """ Test a few typical map_vo use cases """
        # Check things still work if section is missing
        assert map_vo("def") == "def"
        assert map_vo("tst") == "tst"
        assert map_vo("test.vo1-one") == "test.vo1-one"

        # Add config and do mapping tests
        # This first VO name uses all allowed character sets for the long VO name
        core_config.set("vo-map", "test.vo1-one", "tst")
        core_config.set("vo-map", "second.vo", "ts2")

        # Mapping not in config
        assert map_vo("test") == "test"
        # VO in config, but use short name directly
        assert map_vo("tst") == "tst"
        # Test two mappings from config
        assert map_vo("test.vo1-one") == "tst"
        assert map_vo("second.vo") == "ts2"
        # Invalid VO name tests
        # Generate a list of all 1-byte characters and remove ones that we should accept
        test_chars = set([chr(x) for x in range(0, 256)])
        test_chars -= set(ascii_letters)
        test_chars -= set(digits)
        test_chars -= set(['.', '-'])
        for test_chr in test_chars:
            with pytest.raises(RucioException):
                map_vo("bad%s" % test_chr)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Core tracer module
"""

import ipaddress
import json
import logging.handlers
import random
import socket
from typing import TYPE_CHECKING, Any, Union, overload

import stomp
from jsonschema import Draft7Validator, ValidationError, validate

from rucio.common.config import config_get, config_get_int, config_get_list
from rucio.common.exception import InvalidObject, TraceValidationSchemaNotFound
from rucio.common.logging import rucio_log_formatter
from rucio.common.schema.generic import TIME_ENTRY, UUID, IPv4orIPv6
from rucio.core.monitor import MetricManager

if TYPE_CHECKING:
    from datetime import datetime

    from referencing.jsonschema import ObjectSchema, Schema

    from rucio.common.types import TraceSchemaDict

METRICS = MetricManager(module=__name__)

CONFIG_COMMON_LOGLEVEL = getattr(logging, config_get('common', 'loglevel', raise_exception=False, default='DEBUG').upper())

CONFIG_TRACE_LOGLEVEL = getattr(logging, config_get('trace', 'loglevel', raise_exception=False, default='DEBUG').upper())
CONFIG_TRACE_LOGFORMAT = config_get('trace', 'logformat', raise_exception=False, default='%(message)s')
CONFIG_TRACE_TRACEDIR = config_get('trace', 'tracedir', raise_exception=False, default='/var/log/rucio/trace')
CONFIG_TRACE_MAXBYTES = config_get_int('trace', 'maxbytes', raise_exception=False, default=1000000000)
CONFIG_TRACE_BACKUPCOUNT = config_get_int('trace', 'backupCount', raise_exception=False, default=10)

# reset root logger handlers. Otherwise everything from ROTATING_LOGGER will also end up in the apache logs.
logging.getLogger().handlers = []

LOGGER = logging.getLogger('trace')
LOGGER.setLevel(CONFIG_COMMON_LOGLEVEL)

ROTATING_LOGGER = logging.getLogger('trace_buffer')
ROTATING_LOGGER.setLevel(CONFIG_TRACE_LOGLEVEL)

HANDLER = logging.StreamHandler()
FORMATTER = rucio_log_formatter()
HANDLER.setFormatter(FORMATTER)
LOGGER.addHandler(HANDLER)

ROTATING_HANDLER = logging.handlers.RotatingFileHandler(filename='%s/trace' % CONFIG_TRACE_TRACEDIR, maxBytes=CONFIG_TRACE_MAXBYTES, backupCount=CONFIG_TRACE_BACKUPCOUNT)
ROTATING_LOGFORMATTER = logging.Formatter(CONFIG_TRACE_LOGFORMAT)
ROTATING_HANDLER.setFormatter(ROTATING_LOGFORMATTER)
ROTATING_LOGGER.addHandler(ROTATING_HANDLER)

BROKERS_ALIAS, BROKERS_RESOLVED = [], []
try:
    BROKERS_ALIAS = config_get_list('trace', 'brokers')
except:
    raise Exception('Could not load brokers from configuration')

PORT = config_get_int('trace', 'port')
TOPIC = config_get('trace', 'topic')
USERNAME = config_get('trace', 'username')
PASSWORD = config_get('trace', 'password')
VHOST = config_get('trace', 'broker_virtual_host', raise_exception=False)

TOUCH_SCHEMA: 'ObjectSchema' = {
    "description": "touch one or more DIDs",
    "type": "object",
    "properties": {
        "eventType": {"enum": ["touch"]},
        "clientState": {"type": "string"},
        "account": {"type": "string"},
        "scope": {"type": "string"},
        "filename": {"type": "string"},
        "datasetScope": {"type": ["string", "null"]},
        "dataset": {"type": ["string", "null"]},
        "traceTimeentry": TIME_ENTRY,
        "traceTimeentryUnix": {"type": "number"},
        "traceIp": IPv4orIPv6,
        "traceId": UUID,
        "localSite": {"type": "string"},
        "remoteSite": {"type": "string"},
        "usrdn": {"type": "string"},
    },
    "required": ['eventType', 'clientState', 'account', 'traceTimeentry', 'traceTimeentryUnix', 'traceIp', 'traceId']
}

UPLOAD_SCHEMA: 'ObjectSchema' = {
    "description": "upload method",
    "type": "object",
    "properties": {
        "eventType": {"enum": ["upload"]},
        "hostname": {"type": "string"},
        "eventVersion": {"type": "string"},
        "clientState": {"type": "string"},
        "account": {"type": "string"},
        "uuid": UUID,
        "scope": {"type": "string"},
        "datasetScope": {"type": ["string", "null"]},
        "dataset": {"type": ["string", "null"]},
        "remoteSite": {"type": "string"},
        "filesize": {"type": "number"},
        "protocol": {"type": "string"},
        "transferStart": {"type": "number"},
        "transferEnd": {"type": "number"},
        "traceTimeentry": TIME_ENTRY,
        "traceTimeentryUnix": {"type": "number"},
        "traceIp": IPv4orIPv6,
        "traceId": UUID,
        "vo": {"type": "string"},
        "stateReason": {"type": "string"},
        "filename": {"type": "string"},
        "name": {"type": "string"},
        "usrdn": {"type": "string"},
    },
    "required": ['hostname', 'account', 'eventType', 'eventVersion', 'uuid', 'scope', 'dataset',
                 'remoteSite', 'filesize', 'protocol', 'transferStart', 'traceTimeentry', 'traceTimeentryUnix',
                 'traceIp', 'traceId']
}

DOWNLOAD_SCHEMA: 'ObjectSchema' = {
    "description": "download method",
    "type": "object",
    "properties": {
        "eventType": {"enum": ["download"]},
        "hostname": {"type": "string"},
        "eventVersion": {"type": "string"},
        "localSite": {"type": "string"},
        "remoteSite": {"type": "string"},
        "account": {"type": "string"},
        "uuid": UUID,
        "scope": {"type": "string"},
        "filename": {"type": "string"},
        "datasetScope": {"type": ["string", "null"]},
        "dataset": {"type": ["string", "null"]},
        "filesize": {"type": ["number", "null"]},
        "clientState": {"type": "string"},
        "stateReason": {"type": "string"},
        "protocol": {"type": "string"},
        "transferStart": {"type": "number"},
        "transferEnd": {"type": "number"},
        "traceTimeentry": TIME_ENTRY,
        "traceTimeentryUnix": {"type": "number"},
        "traceIp": IPv4orIPv6,
        "traceId": UUID,
        "vo": {"type": "string"},
        "usrdn": {"type": "string"},
        "name": {"type": "string"},
    },
    "required": ['hostname', 'eventType', 'localSite', 'account', 'eventVersion', 'uuid', 'scope',
                 'filename', 'datasetScope', 'dataset', 'filesize', 'clientState', 'stateReason']
}

GET_SCHEMA: 'ObjectSchema' = {
    "description": "get method, mainly sent by pilots",
    "type": "object",
    "properties": {
        "eventType": {"enum": ["get", "get_sm", "sm_get", "get_sm_a", "sm_get_a"]},
        "clientState": {"type": "string"},
        "stateReason": {"type": "string"},
        "url": {"type": ["string", "null"]},
        "vo": {"type": "string"},
        "scope": {"type": "string"},
        "eventVersion": {"type": "string"},
        "remoteSite": {"type": "string"},
        "datasetScope": {"type": "string"},
        "dataset": {"type": "string"},
        "filename": {"type": "string"},
        "name": {"type": "string"},
        "traceTimeentry": TIME_ENTRY,
        "traceTimeentryUnix": {"type": "number"},
        "traceIp": IPv4orIPv6,
        "traceId": UUID,
        "usrdn": {"type": "string"},
    },
    "required": ['eventType', 'localSite', 'eventVersion', 'uuid', 'scope',
                 'filename', 'dataset']
}

PUT_SCHEMA: 'ObjectSchema' = {
    "description": "get method, mainly sent by pilots",
    "type": "object",
    "properties": {
        "eventType": {"enum": ["put_sm", "put_sm_a"]},
        "clientState": {"type": "string"},
        "stateReason": {"type": "string"},
        "url": {"type": ["string", "null"]},
        "vo": {"type": "string"},
        "scope": {"type": "string"},
        "eventVersion": {"type": "string"},
        "remoteSite": {"type": "string"},
        "datasetScope": {"type": "string"},
        "dataset": {"type": "string"},
        "filename": {"type": "string"},
        "name": {"type": "string"},
        "traceTimeentry": TIME_ENTRY,
        "traceTimeentryUnix": {"type": "number"},
        "traceIp": IPv4orIPv6,
        "traceId": UUID,
        "usrdn": {"type": "string"},
        "pq": {"type": "string"},
        "localSite": {"type": "string"}
    },
    "required": ['eventType', 'localSite', 'eventVersion', 'uuid',
                 'filename', 'dataset']
}

SPECIAL_SCHEMA: 'ObjectSchema' = {
    "description": "A special schema to capture most unsupported eventTypes",
    "type": "object",
    "properties": {
        "eventType": {"enum": ["sfo2eos"]},
        "clientState": {"type": "string"},
        "account": {"type": "string"},
        "scope": {"type": "string"},
        "filename": {"type": "string"},
        "datasetScope": {"type": ["string", "null"]},
        "dataset": {"type": ["string", "null"]},
        "traceTimeentry": TIME_ENTRY,
        "traceTimeentryUnix": {"type": "number"},
        "traceIp": IPv4orIPv6,
        "traceId": UUID,
        "localSite": {"type": "string"},
        "remoteSite": {"type": "string"},
        "usrdn": {"type": "string"},
    },
    "required": ['eventType', 'clientState', 'account', 'traceTimeentry', 'traceTimeentryUnix', 'traceIp', 'traceId']
}

SCHEMAS: dict[str, 'Schema'] = {
    'touch': TOUCH_SCHEMA,
    'upload': UPLOAD_SCHEMA,
    'download': DOWNLOAD_SCHEMA,
    'get': GET_SCHEMA,
    'get_sm': GET_SCHEMA,
    'sm_get': GET_SCHEMA,
    'get_sm_a': GET_SCHEMA,
    'sm_get_a': GET_SCHEMA,
    'put': PUT_SCHEMA,
    'put_sm': PUT_SCHEMA,
    'put_sm_a': PUT_SCHEMA,
    'sm_put': PUT_SCHEMA,
    'sm_put_a': PUT_SCHEMA,
    'sfo2eos': SPECIAL_SCHEMA
}

FORMAT_CHECKER = Draft7Validator.FORMAT_CHECKER


@FORMAT_CHECKER.checks(format="ipv4_or_ipv6")
def ip_format_checker(value: str) -> bool:
    """
    Validates IPv4 or IPv6 string values. json schemas can use `ipv4_or_ipv6` as a valid `format` argument
    """
    try:
        ipaddress.ip_address(value)
    except ValueError:
        LOGGER.debug(f"{value} is not a valid IPv4 or IPv6 address and raises an errors upon validation.")
        result = False
    else:
        result = True
    return result


logging.getLogger("stomp").setLevel(logging.CRITICAL)

for broker in BROKERS_ALIAS:
    try:
        addrinfos = socket.getaddrinfo(broker, 0, socket.AF_INET, 0, socket.IPPROTO_TCP)
        BROKERS_RESOLVED = [ai[4][0] for ai in addrinfos]
    except:
        pass

CONNS = []

for broker in BROKERS_RESOLVED:
    CONNS.append(stomp.Connection(host_and_ports=[(broker, PORT)], vhost=VHOST, reconnect_attempts_max=3))


@overload
def date_handler(obj: "datetime") -> str:
    ...


@overload
def date_handler(obj: object) -> object:
    ...


def date_handler(obj: Any) -> Union[str, object]:
    """ format dates to ISO format """
    return obj.isoformat() if hasattr(obj, 'isoformat') else obj


@METRICS.count_it
def trace(payload: dict[str, Any]) -> None:
    """
    Write a trace to the buffer log file and send it to active mq.

    :param payload: Python dictionary with trace report.
    """

    report = json.dumps(payload, default=date_handler)
    ROTATING_LOGGER.debug(report)
    t_conns = CONNS[:]

    try:
        validate_schema(report)
    except InvalidObject as error:
        ROTATING_LOGGER.warning("Problem validating schema: %s" % error)
        LOGGER.warning("Problem validating schema: %s" % error)

    try:
        for i in range(len(t_conns)):
            try:
                conn = random.sample(t_conns, 1)[0]
                if not conn.is_connected():
                    LOGGER.info('reconnect to ' + conn.transport._Transport__host_and_ports[0][0])
                    conn.connect(USERNAME, PASSWORD)
            except stomp.exception.NotConnectedException:
                LOGGER.warning('Could not connect to broker %s, try another one' %
                               conn.transport._Transport__host_and_ports[0][0])
                t_conns.remove(conn)
                continue
            except stomp.exception.ConnectFailedException:
                LOGGER.warning('Could not connect to broker %s, try another one' %
                               conn.transport._Transport__host_and_ports[0][0])
                t_conns.remove(conn)
                continue

        if conn.is_connected:
            conn.send(body=report, destination=TOPIC, headers={'persistent': 'true', 'appversion': 'rucio'})
        else:
            LOGGER.error("Unable to connect to broker. Could not send trace: %s" % report)
    except Exception as error:
        LOGGER.error(error)


def validate_schema(obj: str) -> None:
    """
    Validate object against json schema

    :param obj: The object to validate.

    :raises: InvalidObject, TraceValidationSchemaNotFound
    """
    loaded_obj: 'TraceSchemaDict' = json.loads(obj)

    if loaded_obj and 'eventType' in loaded_obj:
        event_type = loaded_obj['eventType'].lower()
        schema = SCHEMAS.get(event_type)
        if schema is not None:
            try:
                validate(loaded_obj, schema, format_checker=FORMAT_CHECKER)
            except ValidationError as error:
                raise InvalidObject(error)
        else:
            raise TraceValidationSchemaNotFound("Trace schema for eventType %s not found. This event type might not be supported." % event_type)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Increase OAuthRequest.redirect_msg length"""    # noqa: D400, D415

import sqlalchemy as sa
from alembic import context
from alembic.op import alter_column

# Alembic revision identifiers
revision = '30d5206e9cad'
down_revision = 'b0070f3695c8'


def upgrade():
    """Upgrade the database to this revision."""
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        alter_column('oauth_requests', 'redirect_msg', existing_type=sa.String(2048), type_=sa.String(4000), schema=schema)


def downgrade():
    """Downgrade the database to the previous revision."""
    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        schema = context.get_context().version_table_schema if context.get_context().version_table_schema else ''
        alter_column('oauth_requests', 'redirect_msg', existing_type=sa.String(4000), type_=sa.String(2048), schema=schema)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Flask, request

from rucio.common.utils import parse_response
from rucio.gateway.importer import import_data
from rucio.web.rest.flaskapi.authenticated_bp import AuthenticatedBlueprint
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, json_parameters, response_headers


class Import(ErrorHandlingMethodView):
    """ Import data. """

    def post(self):
        """
        ---
        summary: Import data
        description: Import data into rucio
        tags:
            - Import
        requestBody:
          content:
            'application/json':
              schema:
                type: object
                properties:
                  rses:
                    description: Rse data with rse name as key.
                    type: object
                    additionalProperties:
                      x-additionalPropertiesName: rse name
                      type: object
                      properties:
                        rse_type:
                          description: The type of an rse.
                          type: string
                          enum: ['DISK', 'TAPE']
                  distances:
                    description: Distances data with src rse name as key.
                    type: object
                    additionalProperties:
                      x-additionalPropertiesName: src rse
                      description: Distances with dest rse as key.
                      type: object
                      additionalProperties:
                        x-additionalPropertiesName: dest rse
                        description: Distance for two rses.
                        type: object
                        properties:
                          distance:
                            description: The distance between the rses.
                            type: integer
                          ranking:
                            deprecated: true
                            description: Same as distance
                            type: integer
                  accounts:
                    description: Account data.
                    type: array
                    items:
                      description: An account.
                      type: object
                      properties:
                        account:
                          description: The account identifier.
                          type: string
                        email:
                          description: The email of an account.
                          type: string
                        identities:
                          description: The identities associated with an account. Deletes old identities and adds the newly defined ones.
                          type: array
                          items:
                            description: One identity associated with an account.
                            type: object
                            properties:
                              type:
                                description: The type of the identity.
                                type: string
                                enum: ['X509', 'GSS', 'USERPASS', 'SSH', 'SAML', 'OIDC']
                              identity:
                                description: Identifier of the identity.
                                type: string
                              password:
                                description: The password if the type is USERPASS.
                                type: string
        responses:
          201:
            description: OK
            content:
              application/json:
                schema:
                  type: string
                  enum: ['Created']
          401:
            description: Invalid Auth Token
        """
        data = json_parameters(parse_response)
        import_data(data=data, issuer=request.environ.get('issuer'), vo=request.environ.get('vo'))
        return 'Created', 201


def blueprint(with_doc=False):
    bp = AuthenticatedBlueprint('import', __name__, url_prefix='/import')

    import_view = Import.as_view('scope')
    if not with_doc:
        # rule without trailing slash needs to be added before rule with trailing slash
        bp.add_url_rule('', view_func=import_view, methods=['post', ])
    bp.add_url_rule('/', view_func=import_view, methods=['post', ])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation to add the prefix """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint(with_doc=True))
    return doc_app



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cache consumer is a daemon to retrieve rucio cache operation information to synchronize rucio catalog.
"""

import json
import logging
import threading
import time
from traceback import format_exc
from typing import TYPE_CHECKING

import rucio.db.sqla.util
from rucio.common import exception
from rucio.common.logging import formatted_logger, setup_logging
from rucio.common.stomp_utils import ListenerBase, StompConnectionManager
from rucio.common.types import InternalScope, LoggerFunction
from rucio.core.monitor import MetricManager
from rucio.core.rse import get_rse_id
from rucio.core.volatile_replica import add_volatile_replicas, delete_volatile_replicas

if TYPE_CHECKING:
    from types import FrameType

    from stomp.utils import Frame

logging.getLogger("stomp").setLevel(logging.CRITICAL)

METRICS = MetricManager(module=__name__)
GRACEFUL_STOP = threading.Event()
DAEMON_NAME = 'cache-consumer'


class AMQConsumer(ListenerBase):
    """
    class Consumer
    """

    @METRICS.count_it
    def on_message(self, frame: "Frame") -> None:
        """
        on_message
        """
        try:
            msg = json.loads(frame.body)  # type: ignore
            self._logger(logging.DEBUG, 'Message received: %s', msg)
            if isinstance(msg, dict) and 'operation' in msg.keys():
                for f in msg['files']:
                    f['scope'] = InternalScope(f['scope'])
                if 'rse_id' in msg:
                    rse_id = msg['rse_id']
                else:
                    rse_id = get_rse_id(rse=msg['rse'], vo=msg.get('vo', 'def'))

                rse_vo_str = msg['rse']
                if 'vo' in msg and msg['vo'] != 'def':
                    rse_vo_str = f"{rse_vo_str} on {msg['vo']}"
                if msg['operation'] == 'add_replicas':
                    self._logger(logging.INFO, "add_replicas to RSE %s: %s", rse_vo_str, str(msg['files']))
                    add_volatile_replicas(rse_id=rse_id, replicas=msg['files'])
                elif msg['operation'] == 'delete_replicas':
                    self._logger(logging.INFO, "delete_replicas to RSE %s: %s", rse_vo_str, str(msg['files']))
                    delete_volatile_replicas(rse_id=rse_id, replicas=msg['files'])
            else:
                self._logger(logging.DEBUG, 'Check failed: %s %s', isinstance(msg, dict), "operation" in msg.keys())
        except:
            self._logger(logging.ERROR, str(format_exc()))


def consumer(id_: int, num_thread: int = 1, logger: LoggerFunction = logging.log) -> None:
    """
    Main loop to consume messages from the Rucio Cache producer.
    """
    logger(logging.INFO, 'Rucio Cache consumer starting')

    conn_mgr = StompConnectionManager(config_section='messaging-cache', logger=logger)

    logger(logging.INFO, 'consumer started')

    conn_mgr.set_listener_factory('rucio-cache-consumer', AMQConsumer, heartbeats=conn_mgr.config.heartbeats)

    while not GRACEFUL_STOP.is_set():

        conn_mgr.subscribe(id_='rucio-cache-messaging', ack='auto')
        time.sleep(1)

    logger(logging.INFO, 'graceful stop requested')
    conn_mgr.disconnect()
    logger(logging.INFO, 'graceful stop done')


def stop(signum: "int | None" = None, frame: "FrameType | None" = None) -> None:
    """
    Graceful exit.
    """

    GRACEFUL_STOP.set()


def run(num_thread: int = 1) -> None:
    """
    Starts up the rucio cache consumer thread
    """
    setup_logging(process_name=DAEMON_NAME)
    logger = formatted_logger(logging.log, DAEMON_NAME + ' %s')

    if rucio.db.sqla.util.is_old_db():
        raise exception.DatabaseException('Database was not updated, daemon won\'t start')

    logger(logging.INFO, 'starting consumer thread')
    threads = []
    for i in range(num_thread):
        con_thread = threading.Thread(target=consumer, kwargs={'id_': i, 'num_thread': num_thread, 'logger': logger})
        con_thread.start()
        threads.append(con_thread)

    logger(logging.INFO, 'waiting for interrupts')

    while [thread.join(timeout=3.14) for thread in threads if thread.is_alive()]:
        pass



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import shutil
import socket
import tarfile
from collections import OrderedDict
from datetime import datetime, timedelta
from math import asin, cos, radians, sin, sqrt
from pathlib import Path
from tempfile import TemporaryDirectory, TemporaryFile
from typing import IO, TYPE_CHECKING, Any, Optional, Union
from urllib.parse import urlparse

import geoip2.database
import requests
from dogpile.cache.api import NO_VALUE

from rucio.common import utils
from rucio.common.cache import MemcacheRegion
from rucio.common.config import config_get, config_get_bool, config_get_int
from rucio.common.constants import SORTING_ALGORITHMS
from rucio.common.exception import InvalidRSEExpression, SortingAlgorithmNotSupported
from rucio.core.rse_expression_parser import parse_expression

if TYPE_CHECKING:
    from _typeshed import StrPath

    from rucio.common.types import IPDict, ReplicaDict

REGION = MemcacheRegion(expiration_time=900, function_key_generator=utils.my_key_generator)

# This product uses GeoLite data created by MaxMind,
# available from <a href="http://www.maxmind.com">http://www.maxmind.com</a>
GEOIP_DB_EDITION = 'GeoLite2-City'


def extract_file_from_tar_gz(
        archive_file_obj: IO,
        file_name: str,
        destination: 'StrPath'
) -> None:
    """
    Extract one file from the archive and put it at the destination

    archive_fileobj is supposed to be at position 0
    """
    with TemporaryDirectory(prefix=file_name) as tmp_dir:
        tmp_dir = Path(tmp_dir)
        with tarfile.open(fileobj=archive_file_obj, mode='r:gz') as tfile:
            tfile.extractall(path=tmp_dir)  # noqa: S202
            for entry in tfile:
                if entry.name.find(file_name) > -1:
                    print('Will move %s to %s' % (tmp_dir / entry.name, destination))
                    shutil.move(tmp_dir / entry.name, destination)


def __download_geoip_db(destination: 'StrPath') -> None:
    edition_id = GEOIP_DB_EDITION
    download_url = config_get('core', 'geoip_download_url', raise_exception=False, default=None)
    verify_tls = config_get_bool('core', 'geoip_download_verify_tls', raise_exception=False, default=True)
    if not download_url:
        licence_key = config_get('core', 'geoip_licence_key', raise_exception=False, default=None)
        if not licence_key:
            raise Exception('Cannot download GeoIP database: licence key not provided')
        download_url = 'https://download.maxmind.com/app/geoip_download?edition_id=%s&license_key=%s&suffix=tar.gz' % (edition_id, licence_key)

    result = requests.get(download_url, stream=True, verify=verify_tls)
    if result and result.status_code in [200, ]:
        with TemporaryFile() as file_obj:
            for chunk in result.iter_content(8192):
                file_obj.write(chunk)
            file_obj.seek(0)

            extract_file_from_tar_gz(archive_file_obj=file_obj, file_name=f'{edition_id}.mmdb', destination=destination)
    else:
        raise Exception('Cannot download GeoIP database: %s, Code: %s, Error: %s' % (edition_id,
                                                                                     result.status_code,
                                                                                     result.text))


def __geoip_db() -> geoip2.database.Reader:
    db_path = Path(f'/tmp/{GEOIP_DB_EDITION}.mmdb')
    db_expire_delay = timedelta(days=config_get_int('core', 'geoip_expire_delay', raise_exception=False, default=30))

    must_download = False
    if not db_path.is_file():
        print('%s does not exist. Downloading it.' % db_path)
        must_download = True
    elif db_expire_delay and datetime.fromtimestamp(db_path.stat().st_mtime) < datetime.now() - db_expire_delay:
        print('%s is too old. Re-downloading it.' % db_path)
        must_download = True

    if must_download:
        __download_geoip_db(destination=db_path)

    return geoip2.database.Reader(str(db_path))


def __get_lat_long(
        se: str,
        gi: geoip2.database.Reader
) -> tuple[Optional[float], Optional[float]]:
    """
    Get the latitude and longitude on one host using the GeoLite DB
    :param se : A hostname or IP.
    :param gi : A Reader object (geoip2 API).
    """
    try:
        ip = socket.getaddrinfo(se, None)[0][4][0]
        response = gi.city(ip)
        return response.location.latitude, response.location.longitude
    except socket.gaierror as error:
        # Host definitively unknown
        print(error)
    return None, None


def __get_distance(
        se1: str,
        client_location: 'IPDict',
        ignore_error: bool
) -> float:
    """
    Get the distance between 2 host using the GeoLite DB
    :param se1 : A first hostname or IP.
    :param client_location : contains {'ip', 'fqdn', 'site', 'latitude', 'longitude'}
    :ignore_error: Ignore exception when the GeoLite DB cannot be retrieved
    """
    # does not cache ignore_error, str.lower on hostnames/ips is fine
    canonical_parties = list(map(lambda x: str(x).lower(), [se1, client_location['ip'], client_location.get('latitude', ''), client_location.get('longitude', '')]))
    canonical_parties.sort()
    cache_key = f'replica_sorter:__get_distance|site_distance|{"".join(canonical_parties)}'.replace(' ', '.')
    cache_val = REGION.get(cache_key)
    if cache_val is NO_VALUE:
        try:
            gi = __geoip_db()

            lat1, long1 = __get_lat_long(se1, gi)

            lat2 = client_location.get('latitude')
            long2 = client_location.get('longitude')

            if not (lat2 and long2) and client_location['ip'] is not None:
                lat2, long2 = __get_lat_long(client_location['ip'], gi)

            if lat1 and lat2:
                long1, lat1, long2, lat2 = map(radians, [long1, lat1, long2, lat2])  # type: ignore (lat2 and long2 might be None)
                dlon = long2 - long1
                dlat = lat2 - lat1
                cache_val = 6378 * 2 * asin(sqrt(sin(dlat / 2)**2 + cos(lat1) * cos(lat2) * sin(dlon / 2)**2))
                REGION.set(cache_key, cache_val)
                return cache_val
        except Exception as error:
            if not ignore_error:
                raise error
        # One host is on the Moon
        cache_val = 360000
        REGION.set(cache_key, cache_val)
    return cache_val  # type: ignore (cache_val should not be None)


def __download_custom_distance_table() -> None:
    """
    Downloads and parses the custom distance table specified by custom_distance_download_url
    in the config file. Each line of this CSV file should contain a site name, a RSE name,
    and a numerical distance value. Any additional fields are silently ignored.
    """
    db_path = Path('/tmp/rucio_custom_distance_table.csv')
    db_expire_delay = timedelta(days=config_get_int('core', 'custom_distance_expire_delay', raise_exception=False, default=30))

    # check if need to download the file
    must_download = False
    if not db_path.is_file():
        print('%s does not exist. Downloading it.' % db_path)
        must_download = True
    elif db_expire_delay and datetime.fromtimestamp(db_path.stat().st_mtime) < datetime.now() - db_expire_delay:
        print('%s is too old. Re-downloading it.' % db_path)
        must_download = True

    if must_download:
        download_url = config_get('core', 'custom_distance_download_url', raise_exception=False, default=None)
        if download_url is None:
            raise Exception('Cannot download custom distance table: no URL provided')
        result = requests.get(download_url, stream=True, verify=False)
        if result and result.status_code in [200, ]:
            with open(db_path, mode='w') as file_obj:
                file_obj.write(result.text)
        else:
            raise Exception('Cannot download custom distance table: %s, Code: %s, Error: %s' % (download_url,
                                                                                                result.status_code,
                                                                                                result.text))

    # parse the local file and add its contents to REGION
    with open(db_path, mode='r') as f:
        lines = f.readlines()
        for line in lines:
            if line.strip() == "":
                # ignore blank lines
                continue
            bits = line.split(",")
            if len(bits) < 3:
                raise Exception('Custom distance table must have at least 3 values per line')
            # ignore additional fields after first 3 (DUNE has some)
            site = bits[0].strip()
            rse = bits[1].strip()
            distance = float(bits[2].strip())
            if distance < 0.0 or distance > 1.0:
                raise Exception('Distances in custom distance table must be in range 0-1')
            cache_key = f'replica_sorter:__get_distance_custom|site_distance|{rse}|{site}'
            REGION.set(cache_key, distance)


def __get_distance_custom(rse: Union[tuple, str], client_location: 'IPDict') -> float:
    """
    Return the distance from a client to a RSE by looking up in custom distance table
    :param rse: RSE name, or tuple containing replica information with RSE name third
    :param client_location: location dictionary containing {'ip', 'fqdn', 'site', 'latitude', 'longitude'}
    :returns: numerical distance value
    """
    # get RSE name out of tuple if necessary
    if isinstance(rse, tuple) and len(rse) == 4:
        rse = rse[2]
    cache_key = f'replica_sorter:__get_distance_custom|site_distance|{rse}|{client_location["site"]}'
    cache_val = REGION.get(cache_key)
    if not isinstance(cache_val, float):
        # download the table and add all its values to the cache
        __download_custom_distance_table()
        cache_val = REGION.get(cache_key)
        if not isinstance(cache_val, float):
            # assume maximum distance if not specified in table
            cache_val = 1.0
    return cache_val


def site_selector(
        replicas: dict[str, 'ReplicaDict'],
        site: str,
        vo: str
) -> list[str]:
    """
    Return a list of replicas located on one site.
    :param replicas : A dict with RSEs as values and replicas as keys (URIs).
    :param site : The name of the site
    :param vo : The vo within which to search for RSEs
    """
    result = []
    try:
        rses = parse_expression("site=%s" % site, filter_={'vo': vo})
    except InvalidRSEExpression:
        return result
    except Exception:
        return result
    rses = [i['rse'] for i in rses]
    for replica in replicas:
        if replicas[replica] in rses:
            result.append(replica)
    return result


def sort_replicas(
        dictreplica: dict[str, Any],
        client_location: 'IPDict',
        selection: Optional[str] = None
) -> list[str]:
    """
    General sorting method for a dictionary of replicas. Returns the List of replicas.

    :param dictreplica: A dict with replicas as keys (URIs).
    :param client_location: Location dictionary containing {'ip', 'fqdn', 'site', 'latitude', 'longitude'}
    :param selection: the selected sorting algorithm.
    :param default: the default sorting algorithm (random, if not defined).
    :returns: the keys of dictreplica in a sorted list.
    """
    if len(dictreplica) == 0:
        return []

    if not selection:
        selection = 'geoip'
    elif selection not in SORTING_ALGORITHMS:
        raise SortingAlgorithmNotSupported('Sorting algorithm: %s is not supported. Supported protocols: %s' % (selection, SORTING_ALGORITHMS))

    items = [(key, value) for key, value in dictreplica.items()]
    # safety check, TODO: remove if all dictreplica values are 4-tuple with priority as second item
    if isinstance(items[0][1], tuple) and len(items[0][1]) == 4:
        # sort by value[1], which is the priority
        items.sort(key=lambda item: item[1][1])
    dictreplica = OrderedDict(items)

    # all sorts must be stable to preserve the priority (the Python standard sorting functions always are stable)
    if selection == 'geoip':
        replicas = sort_geoip(dictreplica, client_location, ignore_error=config_get_bool('core', 'geoip_ignore_error', raise_exception=False, default=True))
    elif selection == 'custom_table':
        replicas = sort_custom(dictreplica, client_location)
    elif selection == 'random':
        replicas = sort_random(dictreplica)

    return replicas


def sort_random(dictreplica: dict[str, Any]) -> list[str]:
    """
    Return a list of replicas sorted randomly.
    :param dictreplica: A dict with replicas as keys (URIs).
    """

    list_replicas = list(dictreplica.keys())
    random.shuffle(list_replicas)
    return list_replicas


def sort_geoip(
        dictreplica: dict[str, Any],
        client_location: 'IPDict',
        ignore_error: bool = False
) -> list[str]:
    """
    Return a list of replicas sorted by geographical distance to the client IP.
    :param dictreplica: A dict with replicas as keys (URIs).
    :param client_location: Location dictionary containing {'ip', 'fqdn', 'site', 'latitude', 'longitude'}
    :param ignore_error: Ignore exception when the GeoLite DB cannot be retrieved
    """

    def distance(pfn: str) -> float:
        url = urlparse(pfn)
        if url.scheme == 'root':
            # handle root proxy urls: root://10.0.0.1//root://192.168.1.1:1094//dpm/....
            sub_url = urlparse(url.path.lstrip('/'))
            if sub_url.scheme and sub_url.hostname:
                url = sub_url
        return __get_distance(url.hostname, client_location, ignore_error)  # type: ignore (hostname might be None)

    return list(sorted(dictreplica, key=distance))


def sort_custom(
        dictreplica: dict[str, Any],
        client_location: 'IPDict'
) -> list[str]:
    """
    Return a list of replicas sorted according to the custom distance table.
    :param dictreplica: A dict with replicas as keys (URIs).
    :param client_location: Location dictionary containing {'ip', 'fqdn', 'site', 'latitude', 'longitude'}
    :param ignore_error: Ignore exception when the GeoLite DB cannot be retrieved
    """

    def distance(pfn: str) -> float:
        return __get_distance_custom(dictreplica[pfn], client_location)

    return list(sorted(dictreplica, key=distance))



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to manage decommissioning configurations."""

from enum import Enum
from typing import Any

from rucio.common.constants import RseAttr
from rucio.core.rse import add_rse_attribute, get_rse_attribute


class DecommissioningStatus(Enum):
    """Decommissioning status flags."""

    PROCESSING = 'processing'
    DONE = 'done'
    SUSPENDED = 'suspended'


class InvalidStatusName(Exception):
    """Exception for invalid decommissioning status name set from command line."""


def config_to_attr(config: dict[str, Any]) -> str:
    """Form the attribute string from a config dictionary.

    :param config: Decommissioning configuration dictionary.
    :returns: Comma-separated key-value string encoding the configuration.
    """
    attr = f'profile={config["profile"].value}'
    if config.get('move_dest'):
        attr += f',move_dest={config["move_dest"]}'
    attr += f',status={config["status"].value}'

    return attr


def attr_to_config(attr: str) -> dict[str, Any]:
    """Form the config dictionary from an attribute string.

    :param attr: Comma-separated key-value string encoding the configuration.
    :returns: Decommissioning configuration dictionary.
    """
    # The decommission attribute is a comma-separated list of key=value settings
    config: dict[str, Any] = dict(map(lambda s: s.split('='), attr.split(',')))
    if 'status' in config:
        try:
            config['status'] = DecommissioningStatus[config['status'].upper()]
        except KeyError as exc:
            raise InvalidStatusName() from exc
    else:
        config['status'] = DecommissioningStatus.PROCESSING

    return config


def set_status(
    rse_id: str,
    status: DecommissioningStatus
) -> None:
    """Update the decommission attribute of the RSE.

    :param rse_id: RSE ID.
    :param status: RSE decommissioning status.
    """
    config = attr_to_config(get_rse_attribute(rse_id, RseAttr.DECOMMISSION))  # type: ignore (get_rse_attribute could return None)
    config['status'] = status
    # add_rse_attribute can handle updating existing entries too
    add_rse_attribute(rse_id, RseAttr.DECOMMISSION, config_to_attr(config))
